""" Copyright 2025 Russell Fordyce

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from .messaging import tidy_list_str
from .unset import UNSET


def schema_collection_f(collection):
    """ returns a function which accepts value and checks if it's a member of a collection

        pedantically forces bools and 1,0 to be different values
            >>> isinstance(True, int)
            True
            >>> 1.0 in [1,0]
            True
            >>> 0 in {True, False}
            True
        as it's possible to get this condition without knowing when hashing, this also rejects any
        bools or numbers in sets as they could contain True,False literal
            >>> {False, 1, True, 0}
            {False, 1}
            >>> {False, True, 0, 1}
            {False, True}
            >>> {0, 1, True, False}
            {0, 1}
    """
    collection_has_UNSET    = False
    collection_has_0        = False
    collection_has_1        = False
    collection_has_True     = False
    collection_has_False    = False
    collection_has_numerics = False

    for item in collection:
        if item is None:
            continue
        if item is UNSET:  # avoid comparison by ==, which the object rejects
            collection_has_UNSET = True
            continue
        if item is True:
            collection_has_True = True
            continue
        if item is False:
            collection_has_False = True
            continue
        if isinstance(item, (int, float)):  # NOTE complex 1,0 also match, so it's rejected next
            collection_has_numerics = True
            if item == 1:
                collection_has_1 = True
            elif item == 0:
                collection_has_0 = True
            continue  # accept all other numbers too!
        if not isinstance(item, str):
            raise TypeError(f"collection can only contain UNSET,None,bool,int,float,str but got '{type(item)}': {repr(item)}")

    if isinstance(collection, (set, frozenset)):
        if collection_has_True or collection_has_False or collection_has_numerics:
            raise TypeError(
                "True and False literals or numbers in sets are expressly rejected"
                " as (True,False) hash the same as (1,0), consider passing a list or using schema_bool or schema_number_f"
            )
    elif not isinstance(collection, (list, tuple)):
        raise TypeError(f"collection must be a list, tuple, or set, but got {type(collection)}: {tidy_list_str(collection)}")

    def schema_collection_inner(value):
        # FIXME don't repeat error and generate it instead
        # never compare UNSET with ==, which the object rejects
        if value is UNSET:
            if collection_has_UNSET:
                return value
            raise ValueError(f"value must be one of [{','.join(repr(a) for a in collection)}], but got '{repr(value)}'")
        # trap True and False before comparing with 1,0 as they're `==` equal
        if value is True:
            if collection_has_True:
                return value
            raise ValueError(f"value must be one of [{','.join(repr(a) for a in collection)}], but got '{repr(value)}'")
        if value is False:
            if collection_has_False:
                return value
            raise ValueError(f"value must be one of [{','.join(repr(a) for a in collection)}], but got '{repr(value)}'")
        # FUTURE consider splitting int and float too
        # trap 1,0 before comparing with `in`, which will match in True,False
        if value == 1:
            if collection_has_1:
                return value
            raise ValueError(f"value must be one of [{','.join(repr(a) for a in collection)}], but got '{repr(value)}'")
        if value == 0:
            if collection_has_0:
                return value
            raise ValueError(f"value must be one of [{','.join(repr(a) for a in collection)}], but got '{repr(value)}'")
        if value in collection:
            return value
        # don't use `tidy_list_str()` as all keys are relevant
        raise ValueError(f"value must be one of [{','.join(repr(a) for a in collection)}], but got '{repr(value)}'")

    return schema_collection_inner


def schema_number_f(callback=int, limits=None, accept_None=False, accept_UNSET=False):
    # FUTURE consider some checks on callback and limits

    def schema_number_inner(value):
        if value is None:
            if accept_None:
                return value
            raise ValueError(f"expected value, but got {value}")
        if value is UNSET:
            if accept_UNSET:
                return value
            raise ValueError(f"expected value, but got {value}")
        value = callback(value)  # can coerce strings
        # TODO consider allowing a range object as limits, but fail for step setting
        #   though beware that `X in range()` only supports exact integer members
        if limits is not None and (value < limits[0] or value > limits[1]):
            # TODO further enforce limits properties (tuple of len=2 of int,float)
            raise ValueError(f"value must be within {limits}, but got '{value}'")
        return value

    return schema_number_inner


def schema_bool(value):
    # TODO consider coercing some strings
    # if isinstance(value, str):
    #     value = value.lower()
    #     if value == "true":
    #         value = True
    #     if value == "false":
    #         value = False
    if value is False:  # avoid coercing (1,0)
        return False
    if value is True:
        return True
    raise ValueError(f"value must be True or False literal, but got '{type(value)}': {repr(value)}")



""" Copyright 2025 Russell Fordyce

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import re
import threading
from textwrap import dedent as text_dedent
from textwrap import indent as text_indent

import numba
import numpy
import sympy
import sympy.printing.pycode
from sympy.utilities.lambdify import MODULES as lambdify_modules

from .config import (
    DTYPES_FILLER_HINT,
    SYMPY_ATOMS_FP,
)
from .data import dtype_result_guess
from .parsers import SymbolOptional
from .unset import UNSET

from .messaging import warn


def signature_generate(symbols, results, data, dtype_result, config):
    """ generate a signature like
          `Array(int64, 1d, C)(Array(int64, 1d, C))`
        note that Arrays can be named and they begin with the name "array", which
          `repr()` -> `array(int64, 1d, C)`

        refer to Numba types docs and Numba Array(Buffer) classes for more details
          https://numba.readthedocs.io/en/stable/reference/types.html
    """
    # FUTURE support for names (mabye an upstream change to numba)
    #   likely further C-stlye like `void(int32 a[], int64 b)`
    # without names, the dtypes are positional, so ordering must be maintained
    # within logic that could reorder the arguments after fixing the signature!
    # however, when the user calls the Expressive instance,
    # data is passed as kwargs `fn(**data)` to the inner function
    mapper = []

    if len(results) != 1:
        raise RuntimeError("BUG: results symbols should have exactly 1 member: {results}")
    name_result   = next(iter(results.keys()))  # NOTE dict of len==1 if given
    result_passed = bool(name_result in data)  # directly check membership

    names_symbols = list(symbols.keys())
    if not result_passed:
        names_symbols.pop()  # drop the result name (guaranteed to be last symbol in dict)
    for name in names_symbols:  # use symbol ordering, not data ordering
        ref   = data[name]
        field = numba.typeof(ref)  # `numba.types.Array(dtype, dims, layout)` or scalar type
        mapper.append(field)

    # TODO consider warning the user that dummy symbols in the data won't be used
    # NOTE collection only contains symbols which are exclusively a dummy
    # dummy_symbols_in_data = {s.name for s in symbols_dummy if s.name in data}
    # if dummy_symbols_in_data:
    #     warn(f"dummy symbols in data will be ignored: {dummy_symbols_in_data}")
    # TODO warn or raise if not all data names used (+config) [ISSUE 43]
    #   len() is sufficient (KeyError earlier if fewer, but may wrap that too)

    # discover result array dimensions
    # TODO simplify result_passed and dims by using `data[name_result]`->`KeyError` to determine path?
    if result_passed:
        dims = set(data[name].ndim for name in names_symbols if (name != name_result and data[name].ndim != 0))
        ndim_result = data[name_result].ndim
        if dims:  # ignore special case where only single values were passed
            # if len(dims) != 1:
            #     warn(f"unequal data dimensions may result in an error: {dims}")
            if ndim_result not in dims:
                raise ValueError(f"result dimensions (ndim={ndim_result}) do not match inputs: {dims}")
    else:
        dims = set(data[name].ndim for name in names_symbols) - {0}
        if not dims:  # should be detected in `data_cleanup`
            raise RuntimeError("BUG: impossible code path reached, cannot determine result array length from input arrays")
        if len(dims) != 1:
            raise ValueError(f"couldn't determine result dimensions from data, please provide a result array: {dims}")
        ndim_result = dims.pop()

    # now build complete signature for Numba to compile
    # FUTURE consider support for additional dimensions in result
    if not result_passed:
        dtype = getattr(numba.types, str(dtype_result))
        result_ntype = numba.types.Array(dtype, ndim_result, "C")
        mapper.append(result_ntype)

    # range offsets when using threadpool parallelizer
    if config["builder.parallel_model.native_threadpool"]:
        mapper.append(numba.types.int64)  # FUTURE is 2**(32-1) enough values to index any real array?
        mapper.append(numba.types.int64)  # FUTURE might pass data length around as some property too

    return numba.types.void(*mapper), result_passed


def next_free_indexer(symbols, dummy_symbols):
    """ get the next free indexer lexically (idx, idx_a, idx_b...) """
    for char_try in " abcdefghijklmnopqrstuvwxyz":
        indexer = f"idx_{char_try}".rstrip("_ ")  # "idx_ " -> "idx"
        if indexer not in symbols and indexer not in dummy_symbols:  # string compare
            return sympy.Idx(indexer)  # FUTURE indexers might become normal Symbols [ISSUE 183]
    raise ValueError(f"couldn't find a suitable indexer, all ascii lowercase characters used in expr: {symbols}")


def index_vectorized_symbols(symbols, dummy_symbols, indexers, results, data):
    """ discover unindexed Symbols which could become IndexedBase
        returns
            indexers          (also mutated by-ref)
            offset_values     new symbols as {name: IndexedBase}
            indexed_variants  mapper for using .subs() on the existing Symbols
    """
    if indexers:
        raise RuntimeError(f"BUG: tried to re-index Indexed symbols: {indexers}")
    indexer = next_free_indexer(symbols, dummy_symbols)
    indexers.update({indexer: [0, 0]})  # always have 0 spread
    offset_values = {}
    indexed_variants = {}
    for name, symbol in symbols.items():
        if name in dummy_symbols:  # dummy members will redefined later
            # continue  # FUTURE nested Sum() instances probably need/support this
            raise RuntimeError(f"BUG: problematic ref is a Symbol and dummy var: {name}")
        base = sympy.IndexedBase(name)
        try:
            if data[name].ndim == 0:
                continue
        except KeyError as ex:       # might be the result name when not passed in data
            if name not in results:  # this should be impossible
                raise RuntimeError(f"BUG: unknown Symbol '{name}' not in data({data.keys()}): {repr(ex)}")
        offset_values[name]      = base
        indexed_variants[symbol] = sympy.Indexed(base, indexer)

    return indexers, offset_values, indexed_variants


def breakout_Piecewise_to_condition_block(expr, symbols, indexers, results, data):
    """ convert Piecewise (after breakout_Sum_to_loop) into a set of conditional statements

        expected simplification
            Sum(a**x, (x, 1, b))
            Eq(result, Sum(a**x, (x, 1, b)))
            Eq(result, Piecewise((b, Eq(a, 1)), ((a - a**(b + 1))/(1 - a), True)))
            Piecewise((Eq(result, b), Eq(a, 1)), (Eq(result, (a - a**(b + 1))/(1 - a)), True)))
            Piecewise((Eq(result, b[idx]), Eq(a[idx], 1)), (Eq(result, (a[idx] - a[idx]**(b[idx] + 1))/(1 - a[idx])), True)))

        final statements
            if ((a[idx] == 1)): result[idx] = b[idx] ; continue
            if (True): result[idx] = (a[idx] - a[idx]**(b[idx] + 1))/(1 - a[idx]) ; continue

        completed template
            def expressive_wrapper(a, b):
                length = len(a)
                result = numpy.full_like(a, -1, dtype=int64)
                for idx in prange(0, length):
                    if ((a[idx] == 1)): result[idx] = b[idx] ; continue
                    if (True): result[idx] = (a[idx] - a[idx]**(b[idx] + 1))/(1 - a[idx]) ; continue
                return result

        NOTE that the rationale for creating a Piecewise() when simplifying this Sum()
        seems to be avoiding the division by zero at a==1
    """
    if not expr.atoms(sympy.Piecewise):
        raise RuntimeError(f"BUG: expected Piecewise(): {expr}")
    if not expr.atoms(sympy.Eq):  # impossible as added during parse
        raise RuntimeError(f"BUG: expected Eq() wrapper: {expr}")

    # coerce the entire expression and all Piecewise sub-functions instances into a single Piecewise
    # ideally this can avoid some calls and/or duplicate calculations, though it may result in a
    # lengthy series of checks in some cases
    # NOTE this also distributes the outer `Eq(result, ...)` into each Piecewise arg
    # TODO consider if a collection of new functions styled like the sum loops is better
    #   `for index, piecewise_block in enumerate(expr_piecewise.atoms(sympy.Piecewise)):`
    # as-written, this would represent a bug in SymPy as `piecewise_fold()` should consume them all
    expr_piecewise = sympy.piecewise_fold(expr)

    # TODO consider `.piecewise_exclusive()` with or without `skip_nan=True` (then skip filling from caller)
    #   the latter might improve effect on results dtype too (must be Float for NaN?)
    #   and skipping a write might be a little faster
    #   `expr_piecewise = sympy.piecewise_exclusive(expr_piecewise)`
    # FUTURE consider if the order can be optimized in validator (though this is surely very data-dependent)
    #   which may also require `.piecewise_exclusive()` to enable reordering

    count_piecewise = len(expr_piecewise.atoms(sympy.Piecewise))

    # it is possible for `piecewise_fold()` to simplify out all `Piecewise()` when multiple exist
    #   >>> e = parse_expr("Piecewise((b, Eq(a, 1)), (a, True)) + Piecewise((a, Eq(a, 1)), (b, True))")
    #   >>> e
    #   Piecewise((a, Eq(a, 1)), (b, True)) + Piecewise((b, Eq(a, 1)), (a, True))
    #   >>> piecewise_fold(e)
    #   a + b
    # otherwise it should result in exactly 1 combined instance
    if count_piecewise < 1:
        return expr_piecewise, "", indexers, symbols
    if count_piecewise != 1:  # likely represents bug in SymPy as it should always be possible to combine
        raise RuntimeError(f"BUG: failed to coerce all Piecewise into a single call: {expr} -> {expr_piecewise}")
    # TODO consider showing Eq distribution
    # if str(expr_piecewise).count("Eq") < 2:
    #     raise RuntimeError(f"BUG: Eq not distributed to inner blocks: {expr} -> {expr_piecewise}")

    # substitute all symbols for Indexed versions with an indexer
    if not expr.atoms(sympy.Indexed):
        indexers, offset_values, indexed_variants = index_vectorized_symbols(symbols, {}, indexers, results, data)
        expr_piecewise = expr_piecewise.subs(indexed_variants)
        symbols.update(offset_values)  # update for the new expr

    # coerce Idx to Symbol [ISSUE 183]
    expr_piecewise = expr_piecewise.subs({s: sympy.Symbol(s.name) for s in expr_piecewise.atoms(sympy.Idx)})

    conditions = []
    for sub_expr, interval in expr_piecewise.args:
        if not isinstance(sub_expr, sympy.Eq):
            raise RuntimeError(f"BUG: failed to distribute Eq to expression {sub_expr}")
        condition = sympy.printing.pycode(interval).strip()
        condition = f"if ({condition}): {sub_expr.lhs} = {sub_expr.rhs} ; continue"  # embed condition block
        if "not supported in python" in condition.lower() or "\n" in condition:
            raise RuntimeError(f"BUG: failed to transform interval to condition: {interval} -> {condition}")
        conditions.append(condition)

    # TODO warn for insufficient conditions?

    # merge all the condition strings into a single block of text
    # caller is responsible for embedding them sensibly
    block_conditions = "\n".join(conditions)

    return None, block_conditions, indexers, symbols


def breakout_Sum_to_loop(expr, symbols, indexers, results, name_result, data, config):
    """ better take on reducing Sum instances
    """
    if indexers or expr.atoms(sympy.Indexed):
        raise RuntimeError(f"BUG: escaped indexers (should be impossible): {indexers} vs {expr.atoms(sympy.Indexed)}")

    attempt_simplify      = config["translate_simplify.build.sum.try_algebraic_convert"]
    threaded_timeout_warn = config["translate_simplify.build.sum.threaded_timeout_warn"]
    allow_piecewise       = config["translate_simplify.build.sum.allow_piecewise_in_simplification"]

    # collection of `Sum()` instances to work with and replace
    exprs_sums = expr.atoms(sympy.Sum)

    if attempt_simplify:
        # FUTURE `.simplify()` (and `summation()`?) can take an indefinite amount of time, consider
        #  - simple timeout [ISSUE 99] (warn user might optimize, then continue with loop functions?)
        #    consider `fork()` or multiprocessing so `simplify()` can be killed after timeout
        #  - tuning `ratio` arg
        #  - additional custom heuristics for `measure`, perhaps just marking `Sum()`
        if threaded_timeout_warn:
            # FUTURE warn for each `sum_block()` so with several the problematic one can be identified
            def watcher_simplify_warn(e, timeout):
                e.wait(timeout)
                if not e.is_set():
                    warn(
                        f"failed to simplify a Sum() instance in expr after {timeout}s"
                        " perform algebraic expand before .build()ing or set or pass config property"
                        " config['translate_simplify.build.sum.try_algebraic_convert']=False to force "
                        " loop codegen path if operation hangs indefinitely"
                    )

            event_watcher  = threading.Event()
            thread_watcher = threading.Thread(target=watcher_simplify_warn, args=(event_watcher, threaded_timeout_warn), daemon=True)
            thread_watcher.start()

        for sum_block in exprs_sums:
            sum_evaluated = sum_block.replace(sympy.Sum, sympy.summation)
            if not sum_evaluated.atoms(sympy.Sum):  # successfully decomposed `Sum()`
                # attempt to simplify numeric intermediates if any floating-point Atoms are used
                # as the result will be a Float and ideally the precision will remain the same
                #   >>> e = parse_expr("Sum(log(x/b), (x,1,10))")
                #   >>> e.doit()  # more precise, but may require many operations
                #   log(1/b) + log(2/b) + log(3/b) + log(4/b) + log(5/b) + log(6/b) + log(7/b) + log(8/b) + log(9/b) + log(10/b)
                #   >>> e.doit().simplify()      # 10! already exceeds 2**16
                #   10*log(1/b) + log(3628800)
                #   >>> e.doit().simplify().n()  # simplified
                #   10.0*log(1/b) + 15.1044125730755
                # NOTE evaluated functions which simplify to a number become `Float`, so
                #   this won't affect the result dtype
                # FUTURE does this make sense to apply to all inputs?
                #   probably not, but some subset could benefit
                if sum_evaluated.atoms(*SYMPY_ATOMS_FP):
                    # NOTE simplify function instead of method used to ease test mocks
                    sum_evaluated = sympy.simplify(sum_evaluated).n()
                # allow user to avoid replacing Sum with a simplified Piecewise and instead
                # continue to generate a SumLoop for it
                # as the Sum instance will still exist, the faster return below won't trigger
                # specifically this avoids `breakout_Piecewise_to_condition_block()` later
                # in case it's discovered to be problematic for some case(s)
                # NOTE other Sum() instances may still be simplified and replaced if more
                # than one exists in the expr
                if sum_evaluated.atoms(sympy.Piecewise) and not allow_piecewise:
                    continue
                expr = expr.replace(sum_block, sum_evaluated)

        if threaded_timeout_warn:
            event_watcher.set()
            thread_watcher.join()

        if not expr.atoms(sympy.Sum):  # successfully simplified all Sum() instances
            block_sum_functions = ""
            return expr, symbols, indexers, results, block_sum_functions

    # iterate to extract dummy vars and raise for problematic structures
    dummy_symbols = {}
    for sum_block in exprs_sums:
        fn_sum = sum_block.function
        limits = sum_block.limits  # tuple of tuples
        # separate and compare limit features
        dummy_var, limit_start, limit_stop = limits[0]

        # keep a collection of all dummy_symbols
        # TODO can this use `dummy_symbols_split()` result? (mostly avoiding KeyError from data[dummy])
        dummy_symbols[dummy_var.name] = dummy_var

    # identify and prepare to convert all vector/tensor Symbols to IndexedBase
    sum_functions = {}
    indexers, offset_values, indexed_variants = index_vectorized_symbols(symbols, dummy_symbols, indexers, results, data)

    # TODO is it possible for there to be no offset_values?
    #   I think this can only happen if all values are singular and a result array is given
    #   otherwise no result length can be determined
    #   though it is possible for an individual `Sum()` to avoid having any

    # FUTURE support and reorder Sum instances to best order if they're nested (consider recursion too)
    #   trivially, maybe `sort()` where key is the count of Sum() atoms and then lexically?
    #   will this already handle outer dummy vars? probably

    # for each Sum(), create a new custom Function which simply iterates across each dummy in its range
    mapper_new_Sum_functions = {}
    for index, sum_block in enumerate(exprs_sums):
        fn_sum = sum_block.function
        limits = sum_block.limits  # tuple of tuples
        dummy_var, limit_start, limit_stop = limits[0]

        # discover and fill arguments
        args = {s.name: s for s in fn_sum.atoms(sympy.Symbol) - {dummy_var}}
        args = {name: args[name] for name in sorted(args.keys())}  # rebuild in lexical order
        # complain about some edge cases
        set_symbols_inner = set(args.keys()) | {s.name for s in limit_start.atoms(sympy.Symbol)} | {s.name for s in limit_stop.atoms(sympy.Symbol)}
        if not set_symbols_inner:
            warn(f"Sum() has no non-dummy symbols (skipped simplification?): {sum_block}")
        if "rowscalar" in set_symbols_inner:
            raise NotImplementedError("BUG: the name literally 'rowscalar' is reserved in Sum() for now (to mangle or by-ref) [ISSUE 143]")

        # TODO consider crushing sum_block to be a safe name then collection + numbering
        # TODO prefix with 0 for length of exprs_sums
        fn_name = f"SumLoop{index}"

        result_dtype_sum = dtype_result_guess(fn_sum, data)  # returns a type like numpy.foo

        # TODO needs some careful analysis about `stop` vs `stop+1` syntax
        # NOTE Symbols are not replaced by IndexedBase instances yet and frozen as-is here
        T = f"""
        def {fn_name}(start, stop, {", ".join(args.keys())}):
            rowscalar = numpy.{str(result_dtype_sum)}(0)
            for {dummy_var} in range(start, stop+1):
                rowscalar += {fn_sum}
            return rowscalar
        """

        # NOTE start and end can independently be Symbol or numeric
        F = sympy.Function(fn_name)(limit_start, limit_stop, *args.values())
        sum_functions[F] = text_dedent(T)  # maps Function to per-row finalized string

        # create a new Function with indexed args which when embedded will refer to the new loop
        F_indexed = sympy.Function(F.name)(*map(
            lambda a: indexed_variants.get(a, a),
            (limit_start, limit_stop, *args.values())
        ))
        mapper_new_Sum_functions[sum_block] = F_indexed

    # make a single block of string functions to embed as closure(s)
    block_sum_functions = "\n".join(sum_functions.values())

    # rewrite each `Sum()` instance in expr with indexed versions with the new name
    for sum_block, F_new in mapper_new_Sum_functions.items():
        expr = expr.replace(sum_block, F_new)

    # update symbols and the result value
    symbols.update(offset_values)
    indexer = next(iter(indexers.keys()))
    expr = expr.replace(results[name_result], sympy.Indexed(symbols[name_result], indexer))

    return expr, symbols, indexers, results, block_sum_functions


def discover_parallelization(data, config):
    """ determine if build should use `numba.prange` """
    if config["builder.parallel_model.native_threadpool"]:
        return False, True
    if config["builder.parallel_model.allow_prange"] is not UNSET:
        return config["builder.parallel_model.allow_prange"], False
    if max(d.ndim for d in data.values()) <= 1:  # at least shouldn't have broadcasting errors
        return True, False
    # internal threading can ignore some broadcasting errors (and more?) [ISSUE 145]
    return False, False


def template_tidy(T):
    """ tidy up template """
    T = text_dedent(T)
    # ideally these wouldn't be needed, but they're very fast and help let closures be
    # embedded directly in code .. this also prepares for future display via [ISSUE 154]
    T = re.sub(r":\n+", r":\n", T)          # drop newline after new scope
    T = re.sub(r"(\s*)def", r"\n\1def", T)  # ensure all closures have a leading newline
    T = re.sub(r"\n{3,}", "\n\n", T)        # replace all excessive newlines
    T = T.strip()
    return T


def debug_inspect_template(t):
    """ helper function which can be easily mocked to get the final template
        there is no need to mock the return as it's never used, only called

        with unittest.mock.patch("expressive.codegen.debug_inspect_template") as mock_t:
            # do stuff
            template_fn = mock_t.call_args[0][0]  # get the function template argument from helper

        FUTURE constructed templates to be kept by unique Build object helpers [ISSUE 154]
    """
    return t


def debug_inspect_namespace(namespace):
    """ helper for testing and debug, see `debug_inspect_template()` """
    return namespace


def loop_function_template_builder(
    expr,
    symbols,
    indexers,
    results,
    result_passed,
    dtype_result,
    data,
    config,
    extend_names,
    extend_closures,
):
    """ generate environment and code to support the function
         - create namespace
         - fill template
         - exec() to generate the code
         - extract new function
    """

    # parallization features
    allow_prange, use_native_threadpool = discover_parallelization(data, config)
    result_passed = result_passed or use_native_threadpool  # use_native_threadpool always provides an array

    # get ready to build the template (note dict of len==1)
    name_result = next(iter(results.keys()))

    if not expr.atoms(sympy.Sum):
        block_sum_functions = ""
    else:
        # experimental Sum removal
        expr, symbols, indexers, results, block_sum_functions = breakout_Sum_to_loop(
            expr,
            symbols,
            indexers,
            results,
            name_result,
            data,
            config,
        )

    block_conditions = None
    if expr.atoms(sympy.Piecewise):
        # FIXME is this ok to let pass? testing ahoy
        #   it shouldn't be possible to pass a Piecewise directly, so all Sum instances
        #   which simplify into a Piecewise probably completely dissolve ..
        #   however, this might be useful in the future too and some user might have a case
        #   for nesting Sum() in Piecewise() and vice-versa
        if block_sum_functions:
            warn(f"BUG: Sum simplification to Piecewise should transform all Sum instances: {expr}")
        # NOTE this almost-always coerces Symbols to be IndexedBase [ISSUE 181]
        #   however, extremely rare cases may simplify and leave them unchanged
        # TODO can this subvert the initial dtype guess?
        try:
            expr, block_conditions, indexers, symbols = breakout_Piecewise_to_condition_block(
                expr,
                symbols,
                indexers,
                results,
                data,
            )
        except Exception as ex:
            raise RuntimeError(
                "caught Exception breaking Piecewise() into a series of conditional statements.."
                " if this is the result of simplifying a Sum() instance, setting the config option"
                " config['translate_simplify.build.sum.allow_piecewise_in_simplification']=False"
                f" or adjusting the sum limits might allow a successful build: {repr(ex)}"
            )

    # build namespace with everything needed to support the new callable
    # simplified version of sympy.utilities.lambdify._import
    _, _, translations, import_cmds = lambdify_modules["numpy"]
    expr_namespace = {"I": 1j}  # alt `copy.deepcopy(lambdify_modules["numpy"][1])`
    for import_line in import_cmds:
        exec(import_line, expr_namespace)
    # NOTE older SymPy versions may not have any translations at all
    #   while the latest seems to be just `{'Heaviside': 'heaviside'}`
    #   however, this is left as a trivial case for my multi-version build+test tool,
    #   which detects this as a line that is not run and has no other testing
    for sympyname, translation in translations.items():
        expr_namespace[sympyname] = expr_namespace[translation]

    # embed some custom functions
    # FUTURE maintain a collection of these and/or upstream these transform(s)
    # NOTE names here will still be overriden by embed if users have custom functions
    #   due to closures having a higher (highest? closest?) scoping, though this may
    #   change in the future and should be carefully attented, especially for the case
    #   of Numba functions and especially any support for functions without a Python
    #   representation (cffi et al.), though users can apply them before or after
    #   using Expressive on their data..
    if expr is None and not block_conditions:  # comparing expr to None avoids TypeError
        raise RuntimeError("BUG: impossible code path: expr or block_conditions must exist")
    # highlight any new instance of upstream `Mod` support (feels unlikely)
    if (expr is not None and expr.atoms(sympy.Mod)) or (block_conditions and re.match("\bMod\b", block_conditions)):
        if "Mod" in expr_namespace:
            warn("upstream Mod (modulo) support is now available, but may be overriden")
            # TODO consider keeping a copy of expr and/or doing more work if a function might be used
    expr_namespace["Mod"] = numpy.mod  # element-wise modulo (alias of `numpy.remainder()`)

    # determine if range or prange should be used for outer loop
    # NOTE this causes the internal logic to use threads and can result in serious
    #   errors which are lost while still returning success!
    if allow_prange:
        exec("from numba import prange", expr_namespace)

    # construct further template components
    names_symbols = list(symbols.keys())  # avoids mutating symbols
    if use_native_threadpool:  # always include result for native_threadpool
        names_symbols.append("_offset_start")
        names_symbols.append("_offset_end")
    block_args = ", ".join(names_symbols)

    # FUTURE manage this with uneven arrays if implemented [ISSUE 10]
    # NOTE this will use the first discovery it can as arrays all have the same length
    #   or will not pass `data_cleanup()`
    name_size_symbol = None
    if result_passed:  # go ahead and use the passed array if possible
        name_size_symbol = name_result
    elif block_conditions:
        name_size_symbol = next(n for n, s in symbols.items() if isinstance(s, sympy.IndexedBase))
    else:
        atom = sympy.IndexedBase if indexers else sympy.Symbol
        for symbol in expr.rhs.atoms(atom):  # pragma nocover  # set order depends on hash seeding PYTHONHASHSEED
            name_symbol = symbol.name
            # some names definitely can't be used
            #  - result not passed
            #  - extend names definitely can't be used (note extend_names is a collection of SymbolOptional:numberlike)
            #  - single entries in data can't be used
            if name_symbol == name_result:
                continue
            if isinstance(symbol, SymbolOptional):  # FUTURE `and name_symbol not in data:`
                continue
            try:
                if data[name_symbol].ndim == 0:
                    continue
            except KeyError as ex:  # should be impossible without manipulating `._extend_names` property
                warn(f"BUG: Symbol '{name_symbol}' not in data is not SymbolOptional ({type(symbol)}): {repr(ex)}")
                continue
            name_size_symbol = name_symbol
            break
    # condition should have been detected in `data_cleanup()` or `verify_indexed_data_vs_symbols()`
    if name_size_symbol is None:
        raise ValueError("BUG: couldn't determine size of result array, at least one symbol must be an array or pass a result array to fill")

    # prepare values to fill template
    if result_passed or use_native_threadpool:
        block_result = ""  # if the result array is passed in, fill it by-index
    else:                  # without it, dynamically create an array of discovered length
        result_filler = DTYPES_FILLER_HINT[dtype_result]  # KeyError should be impossible
        # FUTURE consider allowing `order='F'` [ISSUE 35]
        block_result = f"{name_result} = numpy.full_like({name_size_symbol}, {result_filler}, dtype={dtype_result}, order='C')"
    block_sum_functions = "\n" + text_indent(block_sum_functions, "    " * 3)  # TODO calculate indent

    # build embedded names block (might be just `"\n\n"`)
    block_extend_names = "\n".join(f"{k.name} = {v}" for k, v in extend_names.items())
    block_extend_names = "\n" + text_indent(block_extend_names, "    " * 3) + "\n"

    # build closures block (might be just `"\n\n"`)
    block_extend_closures = "\n\n".join(extend_closures.values())
    block_extend_closures = "\n" + text_indent(block_extend_closures, "    " * 3) + "\n"

    # construct template
    # FUTURE consider errno or errno-like arg to retrieve extra information from namespace
    if not indexers:
        # FUTURE coerce all Symbols to be Indexed in non-trivial exprs [ISSUE 181]
        # always broadcast as `expressive_inner()` takes `result_array` by-reference after [ISSUE 192]
        T = f"""
        def expressive_inner({block_args}):
            {block_extend_names}
            {block_extend_closures}
            {expr.lhs}[:] = {expr.rhs}
        """
    elif len(indexers) == 1:
        indexer, (start, end) = next(iter(indexers.items()))
        # construct range sequence
        start = -start  # flip start to be positive (no -0 in "normal" Python)
        end   = "length" if end == 0 else f"length - {end}"
        if use_native_threadpool:
            block_range  = "range(_offset_start, _offset_end)"
        else:
            range_fn = "prange" if allow_prange else "range"
            # start = -start  # flip start to be positive (no -0 in "normal" Python)
            # end   = f"length" if end == 0 else f"length - {end}"
            block_range  = f"{range_fn}({start}, {end})"

        if block_conditions is None:
            block_loop_inner = f"{expr.lhs} = {expr.rhs}"
        else:
            block_loop_inner = "\n" + text_indent(block_conditions, "    " * 4)  # TODO calculate indent

        # TODO improve accounting for result LHS in range
        # TODO consider disallowing negative LHS offset, though it feels useful
        # TODO is it better to pass length or duplicate it in the wrapper?
        T = f"""
        def expressive_inner({block_args}):
            {block_extend_names}
            {block_extend_closures}
            {block_sum_functions}
            length = len({name_result})
            for {indexer} in {block_range}:
                {block_loop_inner}
        """
    else:
        # FUTURE consider if it's possible to implement this as nested loops [ISSUE 91]
        raise RuntimeError(f"BUG: indexers must be len 1 when provided (see string_expr_to_sympy): {indexers}")

    T_inner = template_tidy(T)

    # wrapper to handle creating result array when `not result_passed` and more
    if result_passed:
        block_args_outer = block_args
    else:
        # TODO prefer symbols collection over re-parsing string?
        # NOTE result array always passed when `use_native_threadpool` which adds the extra range members
        block_args_outer, test_name_result = block_args.rsplit(", ", 1)
        if test_name_result != name_result or test_name_result == "_offset_end":
            raise RuntimeError(f"BUG: bad block_args line trailing name '{test_name_result}': {block_args}")

    # `expressive_inner()` operates by-reference after [ISSUE 192]
    T = f"""
    def expressive_outer({block_args_outer}):
        {block_result}
        expressive_inner_ref({block_args})
        return {name_result}
    """
    T_outer = template_tidy(T)

    # trivial `lambda t:t` helpers for tests to access template string see also [ISSUE 154]
    debug_inspect_namespace(expr_namespace)
    debug_inspect_template(T_outer)
    debug_inspect_template(T_inner)
    return T_inner, T_outer, expr_namespace


def do_compile(config, signature, T_inner, T_outer, expr_namespace, dtype_result_pre=None):
    # build and extract the inner template (Python function)
    exec(T_inner, expr_namespace)
    fn_inner = expr_namespace["expressive_inner"]

    # pre-compile function as the signature is given
    # FUTURE collect into a class with additional properties (just build time?)
    fn_inner_numba = numba.jit(
        signature,
        nopython=True,  # now the default
        # fastmath=True,  # FUTURE config setting [ISSUE 29]
        parallel=config["builder.parallel_model.numba_parallel"],
    )(fn_inner)

    # now dynamically create wrapper which also returns the result array
    expr_namespace["expressive_inner_ref"] = fn_inner_numba
    exec(T_outer, expr_namespace)
    fn_outer = expr_namespace["expressive_outer"]

    # mangle names for verifying
    # TODO should this be part of `loop_function_template_builder()` instead?
    T_outer_py = T_outer
    T_outer_py = re.sub(r"def expressive_outer\(", "def expressive_outer_py(", T_outer_py)
    T_outer_py = re.sub(r"expressive_inner_ref\(", "expressive_inner(", T_outer_py)

    exec(T_outer_py, expr_namespace)
    fn_py = expr_namespace["expressive_outer_py"]

    return fn_outer, fn_py



""" Copyright 2025 Russell Fordyce

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import sys
from threading import Thread
from queue import Queue
from queue import Empty as EmptyQueueException

import numpy

from .config import DTYPES_FILLER_HINT
from .messaging import warn

JOB_THREAD_EXIT = None


def worker_count_from_cores():
    """ the simple count of cores in a system isn't sufficient to determine the correct
        number of threads, instead, the number of worker cores that can be accessed is a
        far better metric - see official docs
         - https://docs.python.org/3/library/multiprocessing.html#multiprocessing.cpu_count
         - https://docs.python.org/3/library/os.html#os.cpu_count
        NOTE under py3.13+ CPython interpreter is killed on startup if PYTHON_CPU_COUNT or -X arg <=0

        additionally, while it's possible to have more process cores than system cores when a
        user simply sets it to be that way, such a case is probably wrong
          % python3.13 -c 'import os ; print(os.process_cpu_count())'
          12
          % PYTHON_CPU_COUNT=1000000 python3.13 -c 'import os ; print(os.process_cpu_count())'
          1000000
        instead, prefer to directly set ParallelRunner threadcount

        FUTURE consider if this should work with the `numba.set_num_threads()` system for prange too
          does it make sense to use upstream or to try and upstream this function?
    """
    cores_system  = os.cpu_count()  # total system cores count (2*physical cores in basically all amd64 systems)

    try:  # directly accept NUMBA_NUM_THREADS env
        workers = cores_process = int(os.environ["NUMBA_NUM_THREADS"])
        if workers < 1:  # definitely an integer if accepted
            raise OSError(f"NUMBA_NUM_THREADS must be int>=1 if set, but got {workers}")
    except KeyError:
        try:  # go ahead and try 3.13+ way to override `os.process_cpu_count()`
            cores_process = int(os.environ["PYTHON_CPU_COUNT"])
        except KeyError:
            cores_process = os.process_cpu_count() if sys.version_info[:2] >= (3, 13) else len(os.sched_getaffinity(0))
        if cores_process >= cores_system:
            # NOTE == case is normal, while > case is almost-certainly an env error
            workers = max(cores_system - 1, 1)  # prefer to reserve 1 core for system overhead
        else:  # cores_process < cores_system
            workers = cores_process

    if workers < 1:  # shouldn't be possible under py3.13, but might be earlier by setting PYTHON_CPU_COUNT=0
        raise OSError(f"impossible worker count detected: {workers}")
    if cores_system == 1 or cores_process == 1 or cores_process > cores_system:
        warn(
            f"possibly a bad environment: cores_system={cores_system} cores_process={cores_process} "
            f"set env NUMBA_NUM_THREADS before running or dynamically call "
            f"expressive.parallel.parallel_runner.update(threadcount=threadcount) to update the "
            f"threadcount to a reasonable value - using {workers} cores for now"
        )

    return workers


def data_slices(length, count):
    blocksize = length // count + 1
    pos_head = 0
    while pos_head < length:
        pos_next = min(pos_head + blocksize, length)  # don't overrun!
        yield pos_head, pos_next  # NOTE end is noninclusive
        pos_head = pos_next


def worker_parallel(Q_jobs, Q_error):
    """ Thread worker which consumes jobs from Queue """
    while True:  # run forever until program exit (daemon thread) or `ParallelRunner.stop()`
        job = Q_jobs.get()  # blocks, but daemon threads are still killed at program exit
        try:
            if job is JOB_THREAD_EXIT:
                return  # stop called: exit and wait for .join()
            fn, data, (start, end) = job  # just unpack 'em for now
            fn(**data, _offset_start=start, _offset_end=end)
        except Exception as ex:
            Q_error.put(ex)  # save the Exception for caller
        finally:  # always mark job as completed
            Q_jobs.task_done()


class ParallelRunner:
    """ parallel runner object class (should only be one instance, but not enforced) """

    def __init__(self, threadcount="auto"):
        self._set_threadcount(threadcount)
        self._Q_jobs  = Queue()
        self._Q_error = Queue()
        self.threads = []  # unfilled until .start() called to fill and start thread pool
        self._started = False  # wait until an instance is built to create threadpool

    def _set_threadcount(self, threadcount):
        if threadcount == "auto":
            threadcount = worker_count_from_cores()
        elif not isinstance(threadcount, int) or threadcount < 1:
            raise ValueError(f"threadcount must be int>=1, but got {repr(threadcount)}")
        self._threadcount = threadcount

    def start(self):
        """ create and start internal thread pool
            multiple calls will be ignored (only a single instance should exist)
            ideally starting the threads during build avoids adding the start overhead to the runtime
        """
        if not self._started:
            for index in range(self._threadcount):
                t = Thread(target=worker_parallel, args=(self._Q_jobs, self._Q_error), daemon=True)
                self.threads.append(t)
                t.start()
            self._started = True
        return self

    def stop(self):
        """ quit threadpool
            pool must be re-started with .start() if this method is called
            this shouldn't be needed during normal use, but has testing benefits
        """
        # TODO consider adding lock, though really this itself shouldn't need to be threadsafe
        #   Queue instance has a lock (.mutex), but re-using it feels troublesome
        if not self._started:
            return
        self.wait()
        for _ in self.threads:  # exit all threads by passing them a None job
            self._Q_jobs.put(JOB_THREAD_EXIT)
        for t in self.threads:
            t.join()  # thread is now ended
        self.threads.clear()  # no more references exist
        self._started = False

    def wait(self):
        # NOTE unlike threads, `Queue.join()` doesn't close the instance, but only blocks until
        #   `.task_done()` is called for each task which has been added to it (atomic counter)
        self._Q_jobs.join()
        try:
            err = self._Q_error.get_nowait()
        except EmptyQueueException:
            return  # success: no errors from workers
        else:       # at least one worker thread raised an Exception
            raise RuntimeError(f"caught ({self._Q_error.qsize() + 1}) Exception(s) in threads: {repr(err)}") from err

    def update(self, threadcount=None):  # FUTURE opportunity to update other properties
        if threadcount is not None:
            self.stop()
            self._set_threadcount(threadcount)
        return self.start()

    def parallelize(self, E, fn, data, dtype_result, result_passed):
        """ main method when E is called with native_threadpool parallelization model
        """
        # FUTURE consider `numpy.array_split()`?
        #   https://numpy.org/doc/stable/reference/generated/numpy.array_split.html
        #   if splitting verify each `ref.base is None`
        #   https://numpy.org/doc/stable/user/basics.copies.html

        # MUST raise for passing the case
        #  - refers to result array
        #  - offset
        # specifically the result data could become modified out-of-order
        # leading to incorrect results
        # no fix feels likely and Numba should/must handle this case too (especially prange)
        # possibly there are hacks like massive unrolling or discovering some `x*0` exists
        # NOTE relies on _prepare et al. to assure data validity in advance
        #   should work be done there for this?

        # get offsets needed for data chunks
        name_result = next(iter(E._results))
        indexers = E._indexers
        # FIXME no effect without indexers, should they all be adapted when parallel_mode set?
        if len(indexers) != 1:
            raise RuntimeError(f"BUG: expected len(indexers)==1, but got {indexers}")
        _, (_start, _end) = next(iter(indexers.items()))  # surprisingly no need for name
        # if name_result in data and _offset_start != 0 and _offset_end != 0:
        if _start != 0 or _end != 0:
            # raise NotImplementedError(f"can't combine passing result array with indexers: {indexers}")
            raise NotImplementedError(f"non-zero indexers are disallowed (for now): {indexers}")

        # just pass the edges?
        # TODO can caller pass length? (from ._prepare()`?)
        # TODO use combined datalen function [ISSUE 176]
        datalen = 0
        for ref in data.values():
            if ref.ndim == 0:
                continue
            datalen = len(ref)  # NOTE doesn't consider shape yet
            break
        else:
            raise RuntimeError("BUG: couldn't determine data length")

        # actually create the result array if it wasn't passed
        # TODO probably always do this to enable function return to be `numba.void`
        if result_passed:
            arr_result = data[name_result]
        else:
            data = data.copy()  # shallow copy enables including new result array
            filler = DTYPES_FILLER_HINT[dtype_result]
            # FIXME needs to be shape aware (at least improve with full_like and/or raise for mixed shapes)
            arr_result = numpy.full(datalen, filler, dtype=dtype_result)
            data[name_result] = arr_result

        # slight overhead, but hopefully helps case where some thread(s) take twice as long
        # TODO surely there's some good research here
        buckets = self._threadcount * 3
        if buckets > datalen:
            buckets = max(datalen // 3, 1)  # must have at least 1 bucket
            warn(f"very little data passed datalen={datalen}, using buckets={buckets}")

        for data_slice in data_slices(datalen, buckets):
            # job = fn, data, data_slice, counter
            job = fn, data, data_slice
            self._Q_jobs.put_nowait(job)

        # wait for jobs to complete (result array filled)
        # NOTE raises for Exception in worker thread
        self.wait()

        return arr_result  # mirror original API


# FUTURE consider pool or single process for simplification to take place in
#   SymPy simplification may take excessively long or never complete
#   in which case the process can be terminated and building continue
#   with a less-simplified expr
# WORKERS_PARALLEL = {
#     "parallelizer": ParallelRunner(),
# }

# NOTE threads are not created until an Expressive().build() is called
#   using the native_threadpool parallelization model
parallel_runner = ParallelRunner()



""" Copyright 2025 Russell Fordyce

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import re

import numpy
import sympy

from .config import DTYPES_SUPPORTED, SYMPY_ATOMS_FP, SYMPY_ATOMS_PROMOTE
from .messaging import warn


def data_cleanup(data):
    """ verify the incoming data can be used

        currently this expects a dict of numpy arrays

        FUTURE optional other numpy-backed arrays support (ie. Pandas, Polars, etc.)
          for now, users can use the relevant .to_numpy() methods
          also consider .to_records()
    """
    if not data:
        raise ValueError("no data provided")
    if not isinstance(data, dict):
        raise TypeError(f"data must be a dict of NumPy arrays or scalars, but got {type(data)}")

    data_cleaned = {}
    vector_length = {}
    for name, ref in data.items():
        # check name is sensible
        # NOTE that expr parsing removes spaces, which might mash symbols together
        if not isinstance(name, str):
            raise ValueError(f"data names must be strings, but got {type(name)}: {repr(name)}")
        # NOTE `name.isidentifier()` and "\w+" allow some unsuitable valies like "π" and "_"
        if not name.isidentifier() or name.startswith("_") or name.endswith("_") or not re.match(r"^[a-zA-Z\d_]+$", name):
            raise ValueError(f"data names must be valid Python names (identifiers) and Symbols, but got '{name}'")
        # TODO consider warning for keywords `keyword.iskeyword(name)`, but allow some?.. like "del"
        # coerce single python values to 0-dim numpy values
        if isinstance(ref, (int, float, complex)):
            # NOTE chooses widest, signed type for Python values (int64,float64,complex128)
            #   if users want another type, just choose it (ie. `numpy.uint32(10)`)
            # TODO is there a better way to do automatic typing?
            # TODO config option to prefer min type `numpy.array(ref, numpy.min_scalar_type(ref))`
            ref = numpy.array(ref)[()]
        if not isinstance(ref, (numpy.ndarray, numpy.number)):
            raise TypeError(f"data must be a dict of NumPy arrays or scalars, but has member ({name}:{type(ref)})")
        if ref.dtype not in DTYPES_SUPPORTED:
            raise TypeError(f"unsupported dtype ({name}:{ref.dtype})")
        # NOTE single (ndim==0) values have shape==() and `len(array)` raises `TypeError: len() of unsized object`
        if ref.ndim == 0:
            vector_length[name] = 0
        elif ref.ndim == 1:
            vector_length[name] = len(ref)
        else:
            vector_length[name] = len(ref)  # FUTURE further analysis needed for additional dimensions
        data_cleaned[name] = ref

    # compare shapes and warn for mixed dimensions
    shapes = set(ref.shape for ref in data_cleaned.values()) - {()}  # ignore single values (0-dim)
    if len(shapes) > 1 and any(ref.ndim > 1 for ref in data.values()):
        warn(f"mixed dimensions may not broadcast correctly, got shapes={shapes}")

    # FUTURE consider support for uneven input arrays when indexed [ISSUE 10]
    #   specifically offsets can make it so the data does not need the same lengths or
    #   can be padded to be the correct length automatically
    #   however, this also seems ripe for confusion and errors
    vector_lengths = set(vector_length.values())
    if vector_lengths == {0}:
        raise ValueError("only single values passed (ndim=0), no arrays (at least a result array must be passed to determine length)")
    elif len(vector_lengths - {0}) != 1:
        raise ValueError(f"uneven data lengths (must be all equal or 0 (non-vector)): {vector_lengths}")

    return data_cleaned


def dtype_result_guess(expr, data):
    """ attempt to automatically determine the resulting dtype given an expr and data

        this is a backup where the user has not provided a result dtype
        possibly it could support warning for likely wrong dtype

        this is not expected to be a general solution as the problem is open-ended
        and likely depends on the real data

        WARNING this logic assumes the highest bit-width is 64
          larger widths will require rewriting some logic!
          intermediately a user should specify the type, assuming
          a (future) numba really has support for it

        FUTURE consider  `numpy.dtype.alignment`
    """
    # set of dtypes from given data
    dtypes_expr = {c.dtype for c in data.values()}  # set of NumPy types

    # throw out some obviously bad cases
    if not dtypes_expr:
        raise ValueError("no data provided")
    dtypes_unsupported = dtypes_expr - set(DTYPES_SUPPORTED.keys())
    if dtypes_unsupported:
        raise TypeError(f"unsupported dtypes: {dtypes_unsupported}")

    # always return a complex type if present
    if numpy.dtype("complex128") in dtypes_expr or expr.atoms(sympy.I):
        return numpy.dtype("complex128")
    # complex64 is a pair of 32-bit floats, but some types don't cast nicely
    if numpy.dtype("complex64") in dtypes_expr:
        width_noncomplex = max(DTYPES_SUPPORTED[dt] for dt in dtypes_expr if not dt.kind == "c")
        if not width_noncomplex or width_noncomplex <= 32:
            return numpy.dtype("complex64")
        if numpy.dtype("int64") in dtypes_expr or numpy.dtype("uint64") in dtypes_expr:
            # NOTE the default casting for int64+complex64 is complex64, even though it doesn't have enough bits
            return numpy.dtype("complex128")
        if numpy.dtype("float64") not in dtypes_expr:
            raise RuntimeError(f"BUG: expected float64, but got {dtypes_expr}")
        return numpy.dtype("complex128")

    max_bitwidth = max(DTYPES_SUPPORTED[dt] for dt in dtypes_expr)

    # FUTURE support for float128 (does Numba support this?)
    if max_bitwidth > 64:
        raise RuntimeError(f"BUG: max_bitwidth {max_bitwidth}: only complex types exceeding 64 are supported: {dtypes_expr}")

    # now only
    if numpy.dtype("float64") in dtypes_expr:
        return numpy.dtype("float64")
    # promote 32-bit float to 64-bit when greater types are present
    if numpy.dtype("float32") in dtypes_expr:
        if max_bitwidth > 32:
            return numpy.dtype("float64")
        return numpy.dtype("float32")

    # detect structures that make the result logically floating-point
    # TODO perhaps these should be part of a structured attempt to constrain inputs
    #   in addition to being available for guessing resulting type,
    #   even if the constraints are (initially) warns, not hard errors
    # see https://docs.sympy.org/latest/modules/functions/elementary.html
    if (
        expr.atoms(
            *SYMPY_ATOMS_FP
        ) or (
            # discover simple division
            # direct Integers are Rational, but fractional atoms are not Integer
            # additionally, simple divisions will simplify to Integer
            #   >>> parse_expr("4").atoms(Rational), parse_expr("4").atoms(Integer)
            #   ({4}, {4})
            #   >>> parse_expr("4/2").atoms(Rational), parse_expr("4/2").atoms(Integer)
            #   ({2}, {2})
            #   >>> e = "4/2*x + 1/3*y"
            #   >>> parse_expr(e).atoms(Rational) - parse_expr(e).atoms(Integer)
            #   {1/3}
            expr.atoms(sympy.Rational) - expr.atoms(sympy.Integer)
        ) or (
            # detect N/x constructs
            #   >>> srepr(parse_expr("2/x"))
            #   "Mul(Integer(2), Pow(Symbol('x'), Integer(-1)))"
            expr.match(sympy.Pow(sympy.Wild("", properties=[lambda a: a.is_Symbol or a.is_Function]), sympy.Integer(-1)))
        )
    ):
        if max_bitwidth <= 16:  # TODO is this a good assumption?
            return numpy.dtype("float32")
        return numpy.dtype("float64")

    # now pick the largest useful int
    # NOTE constant coefficients should all be Integer (Rational) if reached here

    w_signed   = 0  # NOTE Falsey
    w_unsigned = 0
    for dtype in dtypes_expr:
        if numpy.issubdtype(dtype, numpy.signedinteger):
            w_signed = max(w_signed, DTYPES_SUPPORTED[dtype])
        elif numpy.issubdtype(dtype, numpy.unsignedinteger):
            w_unsigned = max(w_unsigned, DTYPES_SUPPORTED[dtype])
        else:
            raise RuntimeError(f"BUG: failed to determine if {dtype} is a signed or unsigned int (is it a float?)")
    if w_signed and w_unsigned:
        raise TypeError("won't guess dtype for mixed int and uint, must be provided")

    # NOTE without a result array, Numba often chooses widest int64 even for int8,int16 when promoting
    if w_signed and not w_unsigned:
        return numpy.dtype("int64") if (w_signed > 32 or expr.atoms(*SYMPY_ATOMS_PROMOTE)) else numpy.dtype("int32")
    if not w_signed and w_unsigned:
        return numpy.dtype("uint64") if (w_unsigned > 32 or expr.atoms(*SYMPY_ATOMS_PROMOTE)) else numpy.dtype("uint32")

    raise RuntimeError(f"BUG: couldn't determine a good result dtype for {dtypes_expr}")


def get_result_dtype(expr_sympy, results, data, dtype_result=None):
    """ ensure the result datatype matches what's given if any
        use a reasonable guess when not provided explicitly or via result data array
    """
    if dtype_result is not None:                  # string convert
        dtype_result = numpy.dtype(dtype_result)  # idempotent for NumPy types

    if results:
        name_result = next(iter(results.keys()))  # NOTE dict of 1 value
        try:
            dtype_data_result = data[name_result].dtype
        except KeyError:  # name not in in data (not passed: create array later)
            dtype_data_result = None
        else:  # data array contains result for dtype, if expressly provided too, ensure they match
            if dtype_result is None:
                dtype_result = dtype_data_result
            else:
                if dtype_data_result != dtype_result:
                    raise ValueError(f"passed mismatched result array ({dtype_data_result}) and result dtype ({dtype_result})")

    # if dtype_result is still None, guess or raise
    if dtype_result is None:
        dtype_result = dtype_result_guess(expr_sympy, data)

    if dtype_result not in DTYPES_SUPPORTED:
        raise RuntimeError(f"BUG: dtype_result ({dtype_result}) not in DTYPES_SUPPORTED")

    # definitely a supported NumPy type now
    return dtype_result


def verify_indexed_data_vs_symbols(symbols, result_passed, data):
    """ if this instance is indexed, make sure the data makes sense for it
        for example, with "a + b[i]", `a` must be a single value and `b` must be an array

        TODO consider if this should be merged with `signature_generate()`
    """
    names_symbols = list(symbols.keys())
    if not result_passed:
        names_symbols.pop()  # drop the result name (guaranteed to be last symbol in dict)

    for name in names_symbols:
        symbol = symbols[name]
        dims   = data[name].ndim
        if isinstance(symbol, sympy.IndexedBase) and dims == 0:
            raise ValueError(f"'{name}' is indexed, but is a single (ndim={dims}) value in data")
        if isinstance(symbol, sympy.Symbol) and dims > 0:
            title = {1: "array", 2: "matrix"}.get(dims, "tensor")
            raise ValueError(f"'{name}' is not indexed, but passed {title} (ndim={dims}) value in data")



""" Copyright 2025 Russell Fordyce

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from collections import UserDict
from copy import deepcopy
from difflib import get_close_matches

import numpy
import sympy

from .messaging import warn
from .schemas import (
    schema_collection_f,
    schema_number_f,
    schema_bool,
)
from .unset import UNSET


# don't modify this directly, instead write to CONFIG
_CONFIG_DEFAULTS = {
    # translate_simplify.parse
    # there's surely some global parsing choices that make sense for certain applications
    #   should `^` be translated to `**` Pow() or bitwise OR
    #   also warning about this (should there be a warning block?)

    # Sum() features
    # `Sum() is expressly an unevaluated summation
    # set to False to prevent this path, which first tries summation and may do other
    # simplifications on the inner function, always converting to a dedicated
    # function which loops over the given range, which may have symbolic (start,end)
    #
    # many `Sum()` instances can be converted to expressions, potentially avoiding a loop
    # of unknown length, such as here (where the range is not known in advance)
    #   >>> parse_expr("Sum(x, (x, a, b))").replace(Sum,summation)
    #  -a**2/2 + a/2 + b**2/2 + b/2
    # `Sum()` can always be replaced by `summation()`, which may produce a simpler expression
    # or the same or a simpler `Sum()` instance
    "translate_simplify.build.sum.try_algebraic_convert": True,
    # some simplifications result in a Piecewise, this controls if a resulting expr
    # which contains Piecewise is considered a successful simplification
    #   True   to immediately return and let the Piecewise be handled by a later
    #          `breakout_Piecewise_to_condition_block()` call
    #   False  result is ignored and continue to generate SumLoop
    "translate_simplify.build.sum.allow_piecewise_in_simplification": True,
    # warn users after N seconds if sum() simplification is taking an excessive amount of time
    # set the value
    #  - to some positive integer for a timeout in seconds
    #  - False to disable this and never spawn a thread
    #
    # as this spawns a new Thread, some users who are careful about their thread count
    # or are using some model that clashes with them may want to disable this
    # users can also simplify their `Sum()`s before passing them in
    # similarly, Windows users may find Thread generally problematic and wish to disable this
    # TODO this feels like it would be happier in a warnings meta-section
    # TODO this may make sense as part of a general simplifications thread or process
    #   (processes can be killed and benefit from `fork()`)
    "translate_simplify.build.sum.threaded_timeout_warn": 20,  # seconds, yes this is a huge default

    # general data and `._prepare()`-related tunables
    # this might be a home for some memory-related options too?
    #
    # what should happen when extra data names are passed?
    # without any change, calling the function raises
    #   Traceback (most recent call last):
    #     fn_python(**data)
    #   TypeError: expressive_wrapper() got an unexpected keyword argument 'm'
    # valid options
    #   "warn"    warn the user and re-run
    #   "raise"   re-raise TypeError (original behavior)
    #   "ignore"  drop Exception and re-run (useful for users calling many exprs on one data collection)
    # FUTURE consider accepting a callable with some API so users can pass a callback
    "data.runtime.unused_data_callback": "warn",

    # per-instance builder settings
    # some autobuild tunables for users to disable the warnings
    "builder.autobuild.allow_autobuild": False,  # prefer never to autobuild
    "builder.autobuild.usage_warn_nag":  True,   # nag caller via warn() about using .build() instead
    # references specific to parallelization mode
    "builder.parallel_model.numba_parallel":    True,   # longer build times, but may improve performance
    "builder.parallel_model.allow_prange":      UNSET,  # decide based upon data ndim
    "builder.parallel_model.native_threadpool": False,  # FUTURE ignore numba parallelization

    # backend.numba
    # other backends?
    # perhaps .jit args
    #   "fastmath": True,
    #   "parallel": True,
    # needs for like distributing cached files on a network share?.. [ISSUE 24]
}

# mapper for renamed keys which returns a tuple
#  - noted key's new name
#  - new default hint: should be True, False, or a string hint
_RENAMED_KEYS_MAPPER = {
    "builder.allow_autobuild": ("builder.autobuild.allow_autobuild", "effectively the same default"),
}


_SCHEMA_CHECKS = {
    "translate_simplify.build.sum.try_algebraic_convert": schema_bool,
    "translate_simplify.build.sum.allow_piecewise_in_simplification": schema_bool,
    "data.runtime.unused_data_callback": schema_collection_f(["warn", "raise", "ignore"]),
    "translate_simplify.build.sum.threaded_timeout_warn": schema_number_f(float),
    "builder.autobuild.allow_autobuild":        schema_bool,
    "builder.autobuild.usage_warn_nag":         schema_bool,
    "builder.parallel_model.numba_parallel":    schema_bool,
    "builder.parallel_model.allow_prange": schema_collection_f([True, False, UNSET]),
    "builder.parallel_model.native_threadpool": schema_bool,
}


class LegacyConfigMapper:

    def __init__(self, data_ref, constructed_path=""):
        self.data = data_ref
        self.constructed_path = constructed_path

    def __getitem__(self, key):
        constructed = f"{self.constructed_path}.{key}"
        try:
            return self.data[constructed]  # complete key match
        except KeyError:
            self.constructed_path = constructed
            return self

    def __setitem__(self, key, value):
        constructed = f"{self.constructed_path}.{key}"
        if constructed in self.data:  # only set when the key exists
            self.data[constructed] = value
            return
        raise KeyError(f"failed to set {constructed} (from '{key}') to '{value}'")


def try_renamed_keys(key):
    """ see if user is referencing a key that's been renamed
        additionally hint if the default has changed
    """
    try:
        key_new, default_hint = _RENAMED_KEYS_MAPPER[key]
    except KeyError:  # key has not been renamed (normal path)
        return key
    # key has been renamed and might have a new default

    # basic message or embed extra details
    default_hint = "" if default_hint is None else f"({default_hint}) "
    warn(f"renamed key '{key}' is now '{key_new}' {default_hint} and may be removed in a future version")

    return key_new


# NOTE this is an unstable API for now which might change between versions
class ConfigWrapper(UserDict):  # FIXME is this still required to override methods vs subclassing dict?
    """ wrapper for config which can inject warnings and handle the legacy layout
         - key types (must be str)
         - keys flattened (originally a nested dict from v2.2.0)
         - schema checking (FUTURE)
        additionally, user can display all keys with `.display()`
    """
    def __init__(self, config=None):
        self._setup = True  # allow arbitrary setitem during
        # opportunity for users to create/control their own wrapped configs
        if config is None:
            # FUTURE consider a reset feature, though that seems like overkill
            config = deepcopy(_CONFIG_DEFAULTS)
        super().__init__(config)
        # self._config_default = config  # FUTURE consider option to reset
        self._setup = False

    def __getitem__(self, key):  # FUTURE `__getitem__(self, key, /)` when dropping 3.7 support
        # special-case keys used in the legacy nested config
        if key in {"warnings", "translate_simplify", "backend"}:
            warn("all keys must be flattened and dot-separated, legacy key retrieval will be removed in a future version")
            legacy_config = LegacyConfigMapper(self.data, key)
            return legacy_config
        if "." not in key:
            raise ValueError("all keys must be flattened and dot-separated")
        key = try_renamed_keys(key)
        return super().__getitem__(key)

    def __setitem__(self, key, value):
        if not isinstance(key, str):
            raise TypeError(f"keys must be str, but got {type(key)}:{repr(key)}")
        key = try_renamed_keys(key)
        if not self._setup and key not in self.data.keys():
            # try to get a key that's pretty close
            # 0.3 picked because it works for TestConfig.test_bad
            # the most-similar string will be first
            keys_partialmatch = get_close_matches(key, self.data.keys(), n=1, cutoff=0.3)  # list of strings or []
            block_partialmatch = "" if not keys_partialmatch else f"did you mean '{keys_partialmatch[0]}'? "
            raise KeyError(f"unknown key '{key}', {block_partialmatch}enumerate all keys with .display() or see config.py")
        value = _SCHEMA_CHECKS[key](value)  # passthrough, mutate, or raise
        return super().__setitem__(key, value)

    def __delitem__(self, key):
        raise KeyError("ConfigWrapper keys cannot be removed")
        # FUTURE consider allowing for setup `super().__delitem__(key)`

    def clear(self):
        raise KeyError("ConfigWrapper keys cannot be removed")
        # FUTURE consider allowing for setup `super().clear()`

    def display(self):
        """ unstable API """
        keys = sorted(self.data.keys())
        width = max(len(k) for k in keys)  # get width to pad output
        return "\n".join(f"{key.ljust(width)} = {self.data[key]}" for key in keys)


# immediately use a wrapper to warn user about changes from 3.x
CONFIG = ConfigWrapper()


DTYPES_SUPPORTED = {
    # numpy.dtype("bool"):     1,
    numpy.dtype("uint8"):    8,
    numpy.dtype("uint16"):  16,
    numpy.dtype("uint32"):  32,
    numpy.dtype("uint64"):  64,
    numpy.dtype("int8"):     8,
    numpy.dtype("int16"):   16,
    numpy.dtype("int32"):   32,
    numpy.dtype("int64"):   64,
    numpy.dtype("float32"): 32,
    numpy.dtype("float64"): 64,
    # numpy.dtype("float128"): 128,  # not supported in Numba [ISSUE 65]
    numpy.dtype("complex64"):   64,
    numpy.dtype("complex128"): 128,
    # numpy.dtype("complex256"): 256,
}

# determine a sensible fill value when creating a result array
# only called when
#  - using indexing (indexers exists)
#  - result array wasn't passed (whatever content it has is used)
# see also DTYPES_SUPPORTED
DTYPES_FILLER_HINT = {
    # numpy.dtype("bool"):,  # FUTURE (probably fail hard and force filling)
    numpy.dtype("uint8"):  0,
    numpy.dtype("uint16"): 0,
    numpy.dtype("uint32"): 0,
    numpy.dtype("uint64"): 0,
    numpy.dtype("int8"):  -1,
    numpy.dtype("int16"): -1,
    numpy.dtype("int32"): -1,
    numpy.dtype("int64"): -1,
    numpy.dtype("float32"): numpy.nan,
    numpy.dtype("float64"): numpy.nan,
    numpy.dtype("complex64"):  numpy.nan,
    numpy.dtype("complex128"): numpy.nan,
}

# SymPy floating-point Atoms
SYMPY_ATOMS_FP = (
    # straightforward floats
    sympy.Float,
    # trancendental constants
    sympy.pi,
    sympy.E,
    # FUTURE general scipy.constants support
    # common floating-point functions
    sympy.log,
    sympy.exp,
    # sympy.sqrt,  # NOTE simplifies to Pow(..., Rational(1,2))
    # sympy.cbrt,  #   can be found with expr.match(cbrt(Wild('a')))
    # trig functions
    sympy.sin, sympy.asin, sympy.sinh, sympy.asinh,
    sympy.cos, sympy.acos, sympy.cosh, sympy.acosh,
    sympy.tan, sympy.atan, sympy.tanh, sympy.atanh,
    sympy.cot, sympy.acot, sympy.coth, sympy.acoth,
    sympy.sec, sympy.asec, sympy.sech, sympy.asech,
    sympy.csc, sympy.acsc, sympy.csch, sympy.acsch,
    sympy.sinc,
    sympy.atan2,
    # LambertW?  # FUTURE results in float or complex result [ISSUE 107]
)

SYMPY_ATOMS_PROMOTE = (
    sympy.Pow,  # at least promotes int32 -> int64
)



""" Copyright 2025 Russell Fordyce

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import time

import numpy
import sympy

from .messaging import warn

NS10E9 = 10**9  # factor for nanoseconds to second(s)


def verify_cmp(data, expr_sympy, fn_python, fn_compiled, indexers):
    """ check if the compiled and python (pre-jit) functions have the same results
        this helps catch undefined behavior in Numba space, such as log(0)
    """
    # FIXME many magic numbers should be part of config subsystem [ISSUE 29]
    lengths = {k: (len(ref) if ref.ndim == 1 else 1) for k, ref in data.items()}
    lengths_max = max(lengths.values())
    data_names_containing_nan = []
    for name, ref in data.items():
        if numpy.isnan(ref).any():
            data_names_containing_nan.append(name)
    if data_names_containing_nan:
        warn(f"some data in {','.join(data_names_containing_nan)} is NaN")

    time_start = time.process_time_ns()
    result_py = fn_python(**data)
    time_py = time.process_time_ns() - time_start

    time_start = time.process_time_ns()
    result_nb = fn_compiled(**data)
    time_nb = time.process_time_ns() - time_start

    # hint user that using a lot of data
    if (time_py > 10 * NS10E9) and (lengths_max > 2000):
        warn(f"excessive data may be slowing native verify (python:{time_py / NS10E9:.2f}s, compiled:{time_nb}ns) (data lengths {lengths})")

    # check if either ran more than 30 seconds
    if (time_py >= 30 * NS10E9) or (time_nb >= 30 * NS10E9):
        warn(f"verify took a long time python:{time_py / NS10E9:.2f}s, compiled:{time_nb / NS10E9:.2f}s")

    # hint that just NumPy might actually be faster
    if lengths_max >= 1000:
        if time_nb / time_py > 2:  # NumPy is at least twice as fast
            warn(f"compiled function ({time_nb}ns) may be slower than direct NumPy ({time_py}ns) (data lengths {lengths})")

    # symbolics -> Number -> evalf()
    # FUTURE consider collecting Exceptions into a single warning reporting multiple rows
    result_sp = None
    if not indexers and all(d.ndim <= 1 for d in data.values()):  # no indexed values or tensors
        # NOTE numpy.nan are never equal, while sympy.nan are structurally equal, but not symbolically
        #   >>> numpy.nan == numpy.nan
        #   False
        #   >>> sympy.nan == sympy.nan
        #   True
        #   >>> sympy.Eq(numpy.nan, numpy.nan)
        #   False
        #   >>> sympy.Eq(sympy.nan, sympy.nan)
        #   False
        # NOTE numpy.log() handling of negative/zero values is -inf, not a complex value
        # so `sympy.zoo` is converted to `-numpy.inf`, not something like `numpy.complex64(numpy.inf)`
        # however, it might be worth considering `numpy.emath.log()`, which returns the "principle value"
        #   >>> sympy.log(0)
        #   zoo
        #   >>> numpy.log(0)
        #   -inf
        #   >>> numpy.emath.log(0)
        #   -inf
        #   >>> sympy.log(-1)
        #   I*pi
        #   >>> numpy.log(-1)
        #   nan
        #   >>> numpy.emath.log(-1)
        #   3.141592653589793j
        mapper_incomparable_sympy_results = {
            sympy.oo:   numpy.inf,
            sympy.zoo: -numpy.inf,
            sympy.nan:  numpy.nan,
        }
        result_sp = []
        for ref in data.values():
            if ref.ndim == 1:
                length = len(ref)
                break
        else:  # should have been trapped in `data_cleanup(data)`
            dims = {name: ref.ndim for name, ref in data.items()}  # pragma nocover (helper for impossible path)
            raise RuntimeError(f"BUG: no values with ndim==1 passed somehow all single values or tensors: {dims}")
            # length = 1  # FUTURE if allowing single values
        for index in range(length):
            row = {}
            row_nan = False  # track if the row has nan
            for symbols in data.keys():  # needed to handle single values
                value = data[symbols] if data[symbols].ndim == 0 else data[symbols][index]
                if numpy.isnan(value):
                    row_nan = True
                    break
                row[symbols] = value

            if row_nan:  # row is broken, write nan and continue
                result_sp.append(numpy.nan)  # NOTE sympy.nan are equal
                continue  # next row
            # directly use result as `Eq(LHS,RHS)` when no indexers are passed
            r = expr_sympy.rhs.subs(row).evalf()
            # print(f"row build {expr_sympy.rhs} -> {expr_sympy.rhs.subs(row)} -> {r}")  # debug
            if sympy.I in r.atoms():  # "3 * 1j" -> "3*I" -> "3j"
                r = complex(r)
            else:
                r = mapper_incomparable_sympy_results.get(r) or float(r)  # nan is not Falsey, 0->0.0
            result_sp.append(r)

    if indexers:
        indexer, (start, end) = next(iter(indexers.items()))
        start = -start
        end   = -end or None  # `1` to `-1`, `0` to `None`
        result = numpy.allclose(result_py[start:end], result_nb[start:end], equal_nan=True)
    elif result_sp:  # not indexed and no tensors
        result = []  # collection of bool
        for index in range(length):
            value_np, value_py, value_sp = result_nb[index], result_py[index], result_sp[index]
            r1 = numpy.allclose(value_np, value_py, equal_nan=True)
            r2 = numpy.allclose(value_py, value_sp, equal_nan=True)
            r3 = numpy.allclose(value_sp, value_np, equal_nan=True)
            result.append(r1 and r2 and r3)  # (this is a bool)
        result = all(result)  # compact collection into single bool
    else:  # tensor route
        try:
            length = next(len(ref) for ref in data.values() if ref.ndim > 1)
        except StopIteration:  # pragma nocover (impossible path)
            dims = {name: ref.ndim for name, ref in data.items()}
            raise RuntimeError("BUG: used tensor path, but no data had ndim>1: {dims}")
        result = []  # collection of bool
        for index in range(length):
            result.append(numpy.allclose(result_nb[index], result_py[index], equal_nan=True))
        result = all(result)  # compact collection into single bool

    results = {
        "nb": result_nb,
        "py": result_py,
        "sp": result_sp,
    }
    # print("results:")  # debug
    # print(results)

    if not result:  # FUTURE opportunity to hard fail here (via from config?) [ISSUE 29]
        raise RuntimeError(f"not allclose({result}) when comparing between NumPy and compiled function")
    return result, results



""" Copyright 2025 Russell Fordyce

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import re
import warnings
from html import escape as html_escape
from textwrap import dedent as text_dedent


# helpers for warnings
class ExpressiveWarning(RuntimeWarning):
    pass


def warn(msg):
    warnings.warn(msg, ExpressiveWarning)


def tidy_list_str(collection, stringify=None, *, maxlen=5, minwidth=2):
    """ helper for tidying up long collections to display """
    if stringify is None:
        stringify = lambda a: a if isinstance(a, str) else repr(a)  # noqa E731
    if not isinstance(collection, list):
        try:
            collection = sorted(collection)  # coerce to list
        except Exception as ex:
            raise TypeError(f"couldn't coerce collection to list ({type(collection)}) via sorted: {repr(ex)}")
    if len(collection) > maxlen and maxlen > (minwidth * 2):  # setting minwidth is not recommended
        head = ",".join(map(stringify, collection[:maxlen - minwidth]))
        tail = ",".join(map(stringify, collection[-minwidth:]))
        return f"[{head} .. {tail}]"
    return f"[{','.join(map(stringify, collection))}]"


def html_display(expr, string_repr):
    """ dedicated Jupyter/IPython notebook pretty printer method
        this is loaded into an iframe, so mathjax is dynamically acquired too
     in order to render the LaTeX output from SymPy
    """
    # NOTE unstable result for now

    expr_latex = expr._repr_latex_()

    # ensure expr can be displayed properly
    # output wrapped by $$ is the normal output, however, it causes the result to be centered
    # instead, \(expr\) form is preferred which makes the result "inline" and aligned as parent
    expr_latex = re.sub(r"^\$\$?([^\$]+)\$\$?$", r"\(\1\)", expr_latex)
    if not (expr_latex.startswith(r"\(") and expr_latex.endswith(r"\)") and len(expr_latex) >= 5):
        # TODO this should probably be an Exception
        warn(rf"unexpected expr format (should be wrapped in $ -> \(\)): {expr_latex}")
        return string_repr

    # TODO improved templating (though I want to keep deps low)
    #   consider some template engine when available
    # generated as-suggested on mozilla https://developer.mozilla.org/en-US/docs/Web/Security/Subresource_Integrity
    #   `cat src/tex-chtml.js | openssl dgst -sha384 -binary | openssl base64 -A`
    # https://github.com/mathjax/MathJax 3.2.2@227c4fecc0037cef1866d03c64c3af10b685916d
    # see also https://github.com/mathjax/MathJax-src
    # TODO get values from config
    # TODO backend options
    #  - local (embed JS string)
    #  - jsdelivr CDN (suggested and current)
    #  - custom (easily allow orgs to use their own CDN or a completely custom renderer)
    template = """
    <!DOCTYPE html>
    <html>
    <head>
    <script type="text/javascript" id="MathJax-script"
        src="https://cdn.jsdelivr.net/npm/mathjax@3.2.2/es5/tex-chtml.js"
        integrity="sha348-AHAnt9ZhGeHIrydA1Kp1L7FN+2UosbF7RQg6C+9Is/a7kDpQ1684C2iH2VWil6r4"
        crossorigin="anonymous"></script>
    </head>
    <body>
    <ul style="list-style-type:none;padding-left:0;">
    {html_list}
    </ul>
    </body>
    </html>
    """

    # stack entries in unordered list
    collected_values = [
        expr_latex,   # advanced representation
        string_repr,  # FIXME unstable
    ]
    html_list = "\n    ".join(f"<li>{html_escape(a)}</li>" for a in collected_values)

    # fill template
    content = text_dedent(template.format(
        html_list=html_list,
    ))

    return content



""" Copyright 2025 Russell Fordyce

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


class UNSET:

    def __repr__(self):
        return "UNSET"

    def __deepcopy__(self, memo):
        # help maintain singleton as deepcopy will return the same object for `is` equality
        return self

    def __eq__(self, other):
        raise TypeError("compare instances with is, not ==")


# somewhat-dirty clobber name to help prevent new instances from accidentally being created
# NOTE new instances can still be created via `type(UNSET)()`
UNSET = UNSET()



#!/usr/bin/env python3

""" Copyright 2024-2025 Russell Fordyce

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import time
from copy import deepcopy

from .analyze import verify_cmp
from .codegen import (
    do_compile,
    loop_function_template_builder,
    signature_generate,
)
from .config import CONFIG
from .data import (
    data_cleanup,
    get_result_dtype,
    verify_indexed_data_vs_symbols,
)
from .messaging import warn, html_display, tidy_list_str
from .parallel import parallel_runner
from .parsers import (
    parse_inputs,
    parse_parallel_model,
)
from .version import version as __version__  # noqa F401


class Expressive:

    def __init__(
        self,
        expr,
        name_result=None,
        symbols=None,
        *,
        config=None,
        extend=None,
        allow_autobuild=None,
        parallel_model=None,
    ):
        # FUTURE make cleanup optional (arg or config)
        (
            self._expr_sympy,
            self._symbols,
            self._indexers,
            self._results,
            self._symbols_dummy,
            self._extend_names,
            self._extend_closures,
        ) = parse_inputs(expr, symbols, name_result, extend)

        # determine parallelization model
        parallel_model = parse_parallel_model(parallel_model)

        # construct local config
        self._config = deepcopy(CONFIG)
        if config is not None:
            if not isinstance(config, dict):
                raise TypeError(f"config must be a dict, but got {type(config)}")
            self._config.update(config)
        self._config.update(parallel_model)  # always compute parallel mode
        if allow_autobuild is not None:
            self._config["builder.autobuild.allow_autobuild"] = allow_autobuild

        # set up some needed collections
        # self._verifications = {}  # FIXME unstable contents for now
        self.signatures_mapper = {}

    def __str__(self):
        # NOTE unstable result for now
        return f"{type(self).__name__}({self._expr_sympy})"

    def __repr__(self):
        # NOTE unstable result for now
        # FUTURE display some major config settings (but most in a dedicated output)
        # FUTURE consider how to support or use `sympy.srepr()`
        allow_autobuild = self._config["builder.autobuild.allow_autobuild"]
        content = [
            f"build_signatures={len(self.signatures_mapper)}",
            f"allow_autobuild={allow_autobuild}",
        ]
        return f"{str(self)} <{','.join(content)}>"

    def _repr_html_(self):
        """ dedicated Jupyter/IPython notebook pretty printer method
            this is loaded into an iframe, so mathjax is dynamically acquired too
            in order to render the LaTeX output from SymPy
            returns a complete html block or 1-liner string (for error)
        """
        # FUTURE consider display with intermediate template views from builder [ISSUE 152]
        return html_display(
            expr=self._expr_sympy,
            string_repr=repr(self),
        )

    def _prepare(self, data, dtype_result):
        """ prepare before build or __call__ """
        data = data_cleanup(data)
        dtype_result = get_result_dtype(self._expr_sympy, self._results, data, dtype_result)
        signature, result_passed = signature_generate(self._symbols, self._results, data, dtype_result, self._config)
        if self._indexers:  # when indexed, the data shape (array vs single values) matter much more
            verify_indexed_data_vs_symbols(self._symbols, result_passed, data)

        return data, dtype_result, signature, result_passed

    def build(self, data, *, dtype_result=None, verify=None):  # arch target?
        """ compile function and collect it in signatures_mapper """
        dtype_result_pre = dtype_result  # keep a reference to hint about small types if error
        data, dtype_result, signature, result_passed = self._prepare(data, dtype_result)

        if self._config["builder.parallel_model.native_threadpool"]:
            parallel_runner.start()  # begins threadpool (idempotent)

        # automatically set to verify when the array is small
        # only happens for pre-builds as __call__ sets `verify=False` when autobuilding
        if verify is None:
            if self._extend_names or self._extend_closures:  # FIXME interrupts verify for now (keep ref in dict/object?)
                verify = False
            else:
                # approximate max array length (ignores offsets)
                lengths_max = max((len(ref) if ref.ndim == 1 else 1) for ref in data.values())
                if lengths_max <= 50:  # FIXME magic numbers to config subsystem [ISSUE 29]
                    verify = True

        # generate Python function
        T_inner, T_outer, expr_namespace = loop_function_template_builder(
            self._expr_sympy,
            self._symbols,
            self._indexers,
            self._results,
            result_passed,
            dtype_result,
            data,
            self._config,  # FUTURE improve argument ordering
            self._extend_names,
            self._extend_closures,
        )

        fn_outer, fn_py = do_compile(
            self._config,
            signature,
            T_inner,
            T_outer,
            expr_namespace,
            dtype_result_pre,
        )

        if verify:
            result, results = verify_cmp(data, self._expr_sympy, fn_py, fn_outer, self._indexers)
            # self._verifications[signature] = result, results  # unstable contents for now

        self.signatures_mapper[signature] = fn_outer

        # FUTURE does it make sense to return the result(s) from verify?
        #   return self, {"py": result_py, "nb": result_nb, "sp": result_sp}
        # no: probably better to make the result(s) a new property
        return self  # enable dot chaining

    def __call__(self, data, dtype_result=None):
        """ call the relevant compiled function for a particular data collection on it
            if signatures_mapper doesn't have the signature, allow_autobuild can be used
            to create it dynamically, though this loses a lot of the runtime execution speed
            benefits available to users who are able to pre-build for all the data
            signatures they have
        """
        data, dtype_result, signature, result_passed = self._prepare(data, dtype_result)

        try:
            fn = self.signatures_mapper[signature]
        except KeyError as ex:
            if not self._config["builder.autobuild.allow_autobuild"]:
                raise KeyError(f"no matching signature for data: use .build() with representative sample data (or set allow_autobuild=True): {repr(ex)}")
            # FUTURE consider adding autobuild callback for any additional caller housekeeping
            if self._config["builder.autobuild.usage_warn_nag"]:
                time_start = time.process_time()
            self.build(data, dtype_result=dtype_result, verify=False)
            if self._config["builder.autobuild.usage_warn_nag"]:
                time_build = time.process_time() - time_start
                warn(f"autobuild took {time_build:.2f}s of process time, prefer .build(sample_data) in advance if possible")
            try:
                fn = self.signatures_mapper[signature]
            except KeyError:  # pragma nocover - bug path
                raise RuntimeError(f"BUG: failed to match signature after autobuild {len(self.signatures_mapper)} signatures")

        if self._config["builder.parallel_model.native_threadpool"]:
            return parallel_runner.parallelize(self, fn, data, dtype_result, result_passed)

        try:
            return fn(**data)
        except TypeError as ex:
            if "got an unexpected keyword argument" not in str(ex):
                raise ex
            _data = {name: ref for name, ref in data.items() if name in self._symbols}
            if result_passed:
                name_result = next(iter(self._results))
                try:
                    _data[name_result] = data[name_result]
                except KeyError:  # pragma nocover - bug path
                    raise RuntimeError(
                        f"BUG: impossible code path, result array '{name_result}' missing from data"
                        f" something is very wrong with _prepare(), data_cleanup(), or the data object {type(data)}"
                    )
            if self._config["data.runtime.unused_data_callback"] != "ignore":
                _unused_data_names = set(data.keys()) - set(_data.keys())
                msg_err = f"some passed data was not used {tidy_list_str(_unused_data_names)}"
                if self._config["data.runtime.unused_data_callback"] == "raise":
                    raise TypeError(f"{msg_err}: {repr(ex)}")
                warn(f"{msg_err}, set data.runtime.unused_data_callback to modify behavior")
            return fn(**_data)



""" Copyright 2025 Russell Fordyce

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import re
from inspect import getsource as inspect_source_text
from textwrap import dedent as text_dedent

import numba
import numpy
import sympy

from .config import DTYPES_SUPPORTED
from .messaging import warn, tidy_list_str


def symbols_given_cleanup(expr, symbols):
    """ helper for producing a mapping of names to symbols
        if unsure or only partially available, prefer to return less over guessing
        this lets a user pass only what the need and let Expressive figure out
        the rest

        FUTURE make some use of SymPy Assumptions (.assumptions0)
          https://docs.sympy.org/latest/guides/assumptions.html
          https://docs.sympy.org/latest/modules/core.html#sympy.core.basic.Basic.assumptions0
    """
    # all are extracted from expr post-parsing if no Symbols are passed (probably None)
    if not symbols:
        return {}

    types_supported_symbols = (sympy.Symbol, sympy.IndexedBase, sympy.Idx)

    if isinstance(symbols, types_supported_symbols):  # singular Symbol as argument
        return {symbols.name: symbols}

    if isinstance(symbols, (list, tuple, set)):
        for symbol in symbols:
            if not isinstance(symbol, types_supported_symbols):
                raise TypeError(f"symbols must be a collection of SymPy Symbols, but got {type(symbol)}")
        return {s.name: s for s in symbols}

    if isinstance(symbols, dict):
        for name, symbol in symbols.items():
            if not isinstance(name, str):
                raise TypeError(f"all names must be strings (str), but got {type(name)}: {name}")
            if not isinstance(symbol, types_supported_symbols):
                raise TypeError(f"unsupported Symbol {type(symbol)}: {symbol}, expected {types_supported_symbols}")
            if symbol.name != name:
                warn(f"name '{name}' doesn't match symbol.name '{symbol.name}' ({sympy.srepr(symbol)})")
        return symbols

    raise TypeError(f"expected a collection of SymPy Symbols, but got ({type(symbols)})")


class SymbolOptional(sympy.Symbol):
    """ Optional Symbol such as a constant provided by the user via extend which
        should be baked into the function template
        in the future, this might be useful to supply optional args when different
        values are available in the passed data

        for numbers, it would be nice to provide a `sympy.NumberSymbol` wrapper
        however, it's behavior expects instances to be singular, atomic values
        which are defined by a single instance of a custom class

        alternatively, `Dummy` instances are intended to be unique and have custom
        naming logic, such that they display like `f"_{self.name}"`
        that said, perhaps it makes sense to inherit from both
    """
    # TODO should additional guarantees or assumptions be applied here?
    #   they might help or hinder later parsing and simplification(s)
    # is_number = True


def extra_functionality_parse(extend):
    """ split and transform extend into dicts
            extra_functions  {sympy.Function: string function}
            extra_symbols    {SymbolOptional: number-like}

        this is not intended to create an inner system hell, just accelerate some math!
        see also https://en.wikipedia.org/wiki/Greenspun%27s_tenth_rule

        consider enabling a global collection of user functions and constants
        and/or enabling `import` statements
    """
    if extend is None:  # no extras, but still provide dicts to ease later `locals_dict.update()`
        return {}, {}
    if not isinstance(extend, dict):
        raise TypeError(f"extend must be a dict if given, but got {type(extend)}")

    # aggressively verify keys before doing any other work
    # the first logic here allowed both str and symbols and deduplicated them, but it seemed needlessly complex
    keys_nonstr  = []
    keys_badname = []
    for key in extend.keys():
        if not isinstance(key, str):
            keys_nonstr.append(key)
            continue  # avoid AttributeError from missing string methods next
        if not key.isidentifier() or key.startswith("_") or key.endswith("_") or not re.match(r"^[a-zA-Z\d_]+$", key):
            keys_badname.append(key)
    if keys_nonstr:
        raise ValueError(f"extend keys must be str, but passed {len(keys_nonstr)} nonstr keys: {tidy_list_str(keys_nonstr)}")
    if keys_badname:
        raise ValueError(
            f"extend keys must be valid Python names without a leading or trailing underscore '_',"
            f" but passed {len(keys_badname)} invalid keys: {tidy_list_str(keys_badname)}"
        )

    # extend_imports  = []  # list of string imports  # TODO consider enabling additional imports
    extend_names    = {}  # {SymbolOptional: number-like}
    extend_closures = {}  # {sympy.Function: string function}

    for key, value in extend.items():
        # TODO prefer lazy message building over immediately for every pair..
        # TODO better show closure value, though strings won't have a length limit and might make huge errors..
        # msg_err = f"expected mapping{{(str,NumberSymbol,Function):(callable,numeric)}}, but got {key}:{type(value)}"

        # determine if value is a single number
        if isinstance(value, (int, float, complex, numpy.number, numpy.ndarray)):
            if hasattr(value, "ndim") and value.ndim != 0:
                raise ValueError(f"extend[{key}] expected ndim==0, but has ndim=={value.ndim}")
            if hasattr(value, "dtype") and value.dtype not in DTYPES_SUPPORTED:
                raise TypeError(f"extend[{key}] unsupported dtype '{value.dtype}'")
            key = SymbolOptional(key)
            value = numpy.array(value)[()]  # TODO use data to improve resulting type
            extend_names[key] = value
            continue

        # coerce functions to embeddable strings (for now?)
        # TODO support for Expressive instances, may require Build instances [ISSUE 154]
        # TODO consider better supporting prebuilt Numba callables (generate and compare signature?)
        #   for now, this instead attempts to get the `.py_func` property from them and raises
        #   if it's not available (probably `inspect.getsource as inspect_source_text` -> `OSError`)
        #   maybe also relevant, consider `from inspect import signature as inspect_signature`
        # TODO improve any later namespace errors related to these, at least suggesting
        #   that they only use from supported imports (numpy et al.)
        #   keeping and reporting the string collection of imports might be sufficient
        # TODO consider str->lambda support, even if they're then transformed into function definitions
        if callable(value):  # try to coerce function to its full text
            if isinstance(value, numba.core.dispatcher.Dispatcher):  # only `numba.core.registry.CPUDispatcher`?
                count_signatures = len(value.signatures)  # FIXME is it possible for this to raise?
                if count_signatures >= 1:  # this is a pre-built Numba function!
                    # WARNING there may not actually be a matching signature for its later use
                    # FUTURE allow direct inclusion
                    warn(f"extend[{key}] prebuilt Numba function (.signatures={count_signatures}) will be coerced to a closure for now")
                value = value.py_func  # FIXME is this somewhat evil?
            value = inspect_source_text(value)  # can raise TypeError (builtins) or OSError (no source text)
            # continue to str path
            # NOTE that the function string may include any amount of leading spaces as-written in the code

        if isinstance(value, str):
            value = text_dedent(value)  # strips all leading spaces
            if not value.startswith("def "):
                raise ValueError(f"extend[{key}] unsable value, expected function 'def' prefix: {value[:50]}")
            name_fn = re.match(r"def ([^\(]+)\(", value).group(1)  # TODO consider AST parse (both?)
            if name_fn != key:
                # TODO consider warning and replacing instead of raising
                #   enforce naming like key as above if changing
                raise ValueError(f"extend[{key}] doesn't match function value's name '{name_fn}'")
            key = sympy.Function(key)
            extend_closures[key] = value
            continue

        raise ValueError(f"extend[{key}] expected number or function-like, but got {type(value)}")

    return extend_names, extend_closures


def dummy_symbols_split(expr, symbols):
    """ extract names that are exclusively used as dummy variables
        for example in `Sum(x, (x, 1, y))`, `x` is not needed or used from data
        and replaced by some function of `y`

        only symbols which are exclusively a dummy will be split out
    """
    dummy_symbols = {}

    # collect dummy values from `Sum()` limits
    # TODO move verification to an earlier step in Expressive.__init__()
    for sum_block in expr.atoms(sympy.Sum):
        if expr.atoms(sympy.Indexed):
            raise NotImplementedError(f"mixing indexing and Sum is not (yet) supported: {expr.atoms(sympy.Indexed)}")
        fn_sum = sum_block.function
        limits = sum_block.limits  # tuple of tuples
        # easy validity checks
        if len(limits) != 1:
            raise NotImplementedError(f"only exactly 1 Sum() limits is supported for now: {limits}")
        if fn_sum.atoms(sympy.Sum):
            raise NotImplementedError(f"nested Sum instances not yet supported: {sum_block} wraps {fn_sum}")
        # entire expr is checked for IndexedBase for now
        # if fn_sum.atoms(sympy.Indexed):  # TODO consider walrus operator and including IndexedBase
        #     raise NotImplementedError(f"indexed Sum are not yet supported: {sum_block} -> {fn_sum.atoms(sympy.Indexed)}")

        # separate and compare limit features
        dummy_var, limit_start, limit_stop = limits[0]
        if not isinstance(limit_start, (sympy.Symbol, sympy.Integer)):
            raise TypeError(f"unsupported type for limit start '{limit_start}': {type(limit_start)}")
        if not isinstance(limit_stop, (sympy.Symbol, sympy.Integer)):
            raise TypeError(f"unsupported type for limit stop '{limit_stop}': {type(limit_stop)}")
        if isinstance(limit_start, sympy.Integer) and isinstance(limit_stop, sympy.Integer):
            if limit_start > limit_stop:  # TODO consider support for negative sum range (`{Sum((_,a,b)):-Sum((_,b,a))}`?)
                raise ValueError(f"fixed Sum() limits start({limit_start}) > stop({limit_stop}) represents a zero or negative range")
            # if limit_start == limit_stop:  # FUTURE simplifier can directly convert this to fn_sum (all fixed limits?)
            #     warn(f"fixed Sum() limits start({limit_start}),stop({limit_stop}) can be eliminated")

        # keep a collection of all dummy_symbols
        # TODO can this use `dummy_symbols_split()` result? (mostly avoiding KeyError from data[dummy])
        dummy_symbols[dummy_var.name] = dummy_var

    # TODO rename dummy_symbols so it's clear they're all in some way optional to provide (or never called?)
    #   plausibly it makes sense to smooth making them optional arguments
    for name, sym in symbols.items():
        if isinstance(sym, SymbolOptional) and name not in dummy_symbols:
            dummy_symbols[name] = sym

    # remove dummy symbols from args collection
    # NOTE symbols collection is mutated here (by-ref) in addition to being returned
    for name in dummy_symbols.keys():
        try:
            del symbols[name]
        except KeyError as ex:  # pragma nocover impossible/bug path
            raise RuntimeError(f"BUG: dummy var '{name}' missing during split from symbols collection: {repr(ex)}")

    return symbols, dummy_symbols


def string_expr_cleanup(expr_string):
    """ a few rounds of basic cleanup to ease usage
        equality is transformed to Eq() form
            `LHS = RHS` -> `LHS==RHS` -> `Eq(LHS,RHS)`
    """
    # FUTURE consider if these can or should use the SynPy transformation system
    #   https://docs.sympy.org/latest/modules/parsing.html#parsing-transformations-reference
    if not isinstance(expr_string, str):
        raise ValueError("expr must be a string")

    # FUTURE consider accepting/allowing Piecewise inputs directly
    #   trivial here perhaps as `and "Piecewise" not in expr_string` in addition to <>
    #   however, additional robustness is needed especially to support additional `==`/`Eq()` instances
    #   it's also possible (though inadvisable) to hack the expr in by creating a simpler expr
    #   with the same symbol names, then clobbering instance._expr_sympy with the Piecewise
    # FUTURE also fix bad casing of Piecewise
    #   `expr_string = re.sub(r"\b[Pp]iece[Ww]ise\b", r"Piecewise", expr_string)`
    #   or more powerfully .lower() -> \bregex\b -> replace by-position (as IGNORECASE flag is not sufficient)
    if "piecewise" in expr_string.lower():
        raise ValueError("Piecewise is only supported as a result of Sum() simplification for now")
    if "<" in expr_string or ">" in expr_string:
        raise ValueError("inequality is not supported")

    # allow the name mod, but transform mod->Mod when it's used as a Function
    # `Mod()` finally becomes `numpy.mod()` in the builder, which is probably what users
    # expect to happen .. this shouldn't break any existing uses of the name literally "mod"
    # as it couldn't be used as a Function yet
    expr_string = re.sub(r"\bmod\s*\(", r"Mod(", expr_string)

    # discard all whitespace to ease string processing
    # this will make (invalid) collections like "a b c" into "abc", not "a*b*c"
    # as a user might conceivably expect, though this does not claim to ..but
    # if so this will be later highlighted when the data is not valid as no
    # "abc" symbol name exists in the given data unless some very special
    # circumstance arises
    # TODO raise or warn for mashing names together [ISSUE 101]
    #   "a cos(b)" -> "acos(b)"
    expr_string = re.sub(r"\s+", r"", expr_string)  # expr_string.replace(" ", "")

    # coerce runs of "=" into exactly "=="
    # ideally only (0,1,2) exist, but let users be really excited ==== for now
    expr_string = re.sub(r"=+", "==", expr_string)
    count_equalities = expr_string.count("=") // 2
    if count_equalities == 1:
        lhs, rhs = expr_string.split("==")  # ValueError if doesn't unpack exactly
        # recurse for each half, then rejoin 'em
        lhs = string_expr_cleanup(lhs)
        rhs = string_expr_cleanup(rhs)
        return f"Eq({lhs}, {rhs})"
    elif count_equalities > 1:  # not exactly 0 or 1 (==)
        raise SyntaxError(f"only 1 equivalence (==) can be provided, but parsed {count_equalities}: {expr_string}")

    # user probably meant Pow() not bitwise XOR
    # TODO add to warning subsystem `if "^" in expr_string:`
    # TODO allow configuring this warning too [ISSUE 29]
    expr_string = expr_string.replace("^", "**")

    # multiplication cleanup blocks
    # SymPy expects symbols to be separated from Numbers for multiplication
    #   ie. "5x+7" -> "5*x+7"
    # however, care needs to be taken to avoid splitting symbols and functions
    # which contain a number, like `t3`, `log2()`, etc.

    # clean up the case where trailing number isn't multiplied like it should be
    #   ie. "(a+b)2" -> "(a+b)*2"
    # TODO add to warning subsystem as it feels like an unusual style [ISSUE 29]
    #   and could easily be a typo like missing Pow '^' or other operator which
    #   would otherwise be silently "fixed"
    # NOTE SymPy might handle this case without the additional '*' after 1.12
    expr_string = re.sub(r"(\))(\d+)", r"\1*\2", expr_string)

    # consider matches where a number appears directly after
    #   start of string | basic operators "+-*/" | open parenthesis
    # and directly before a case where
    #   new string starts (symbol or function)
    #   new parentheses block starts "3(a+b)" -> "3*(a+b)"
    # likely this could be better tokenized by Python AST or SymPy itself
    expr_string = re.sub(r"(^|[\+\-\*\/]|\()(\d+)([a-zA-Z]|\()", r"\1\2*\3", expr_string)

    # make sure there's something left after parsing
    #  (ie. user didn't just pass " " or something went badly ary above)
    if not expr_string:
        raise ValueError("no content after cleanup")

    return expr_string


def get_or_create_symbol(dict_search, name, symbol_cls):
    """ helper like `dict.get(name, sym(name))` which checks matches are the expected type
        this is useful specifically to ensure passed Symbols are more specific types
            IndexedBase
            Idx
        otherwise they can't be used later
    """
    try:
        value = dict_search[name]
    except KeyError:
        return symbol_cls(name)  # directly make one from the name
    if not isinstance(value, symbol_cls):  # name in dict, but wrong type!
        raise TypeError(f"{name} should be type {symbol_cls}, but got {type(value)}")
    return value


def string_expr_indexing_offsets(expr_string, symbols):
    """ detect and manage relative offsets
        returns tuple with
         - offset values like `symbols` mapping {name:Symbol}
         - range the index can be (inclusive)
        symbols will be used if they have the same name as discovered values
        raising if the name is the same, but the Symbol type is wrong
        (refer to get_or_create_symbol)

        for example, given
            a[i+1] + b[i-1]
        this returns like
            offset_values {
                "a": IndexedBase("a")
                "b": IndexedBase("b")
                "i": Idx("i")
            }
            offset_ranges {
                Idx("i"): [-1, 1],
            }
    """
    # FUTURE handle advanced relative indexing logic [ISSUE 11]
    # FUTURE consider if multiple Idx can generate deeper loops
    offset_values = {}
    offset_ranges = {}  # spread amongst offsets as name:[min,max]
    for chunk in re.findall(r"(\w+)\[(.+?)\]", expr_string):
        base, indexing_block = chunk
        indexer = str(sympy.parse_expr(indexing_block).free_symbols.pop())
        try:  # extract the offset amount ie. x[i-1] is -1
            offset = sympy.parse_expr(indexing_block).atoms(sympy.Number).pop()
        except KeyError:
            offset = 0  # no offset like x[i]
        offset_values[base]    = get_or_create_symbol(symbols, base, sympy.IndexedBase)
        offset_values[indexer] = get_or_create_symbol(symbols, indexer, sympy.Idx)
        # now update the spread for the offset
        indexer = offset_values[indexer]  # use Idx ref directly, not name
        spread = offset_ranges.get(indexer, [0, 0])  # start fresh if missing
        spread[0] = min(spread[0], offset)
        spread[1] = max(spread[1], offset)
        offset_ranges[indexer] = spread  # creates if new

    # really make sure there is exactly zero or one indexing Symbols Idx
    if len(offset_ranges) > 1:
        raise ValueError(f"only a single Idx is supported, but got: {offset_ranges}")

    return offset_values, offset_ranges


def indexed_offsets_from_expr(expr):
    """ parse indexed offset features from a SymPy expr

        parallels `string_expr_indexing_offsets()`, though this expects
        the caller to ensure any symbols are present in expr before calling
    """
    if not isinstance(expr, (sympy.core.expr.Expr, sympy.core.relational.Equality)):
        raise RuntimeError(f"BUG: expected SymPy Expr or Equality, but got {type(expr)}")

    offset_values = {}
    offset_ranges = {}  # spread amongst offsets as name:[min,max]
    for block in expr.atoms(sympy.Indexed):
        base    = block.atoms(sympy.IndexedBase)
        indexer = block.atoms(sympy.Idx)
        if len(base) != 1:
            raise ValueError(f"multiple or nested IndexedBase: {block}")
        if len(indexer) != 1:
            raise ValueError(f"indexer must be a single Idx, but got {block}")
        base    = base.pop()  # exactly 1 value exists
        indexer = indexer.pop()
        # now calculate the offset
        offset = (block.atoms(sympy.Rational, sympy.Float) or {0})  # ideally Integer or empytset{}->{0}
        if offset != (block.atoms(sympy.Integer) or {0}):  # error for fractional indicies
            raise ValueError(f"expected a single Integer (or nothing: 0) as the offset, but parsed {block}")
        offset_values[base]    = base
        offset_values[indexer] = indexer
        offset = offset.pop()  # {N} -> N
        spread = offset_ranges.get(indexer, [0, 0])  # start fresh if missing (alt. defaultdict(lambda))
        spread[0] = min(spread[0], offset)
        spread[1] = max(spread[1], offset)
        offset_ranges[indexer] = spread  # creates entry if this is the first instance

    if len(offset_ranges) > 1:
        raise ValueError(f"only a single Idx is supported, but got: {offset_ranges}")

    return offset_values, offset_ranges


def string_expr_to_sympy(expr_string, name_result=None, symbols=None, extend_names=None, extend_closures=None):
    """ parse string to a SymPy expression
        this is largely support logic to help sympy.parse_expr()
         - support for indexing Symbols via IndexBase[Idx]
         - helps make symbol reference collections consistent before and after parsing
           ie. `reference in e.atoms(IndexedBase)` or `foo is atom` are True

        note that `parse_expr()` creates new symbols for any un-named values

        collections of Symbols are returned as dicts mapping {name:Symbol},
        even if there is only a single Symbol
        while any indexer (Idx) is returned as a mapping of
          {Idx:[low index,high index]}
        so the templated loop won't over or under-run its array indices

        FUTURE work with transformation system over regex hackery where possible
    """
    if symbols is None:
        symbols = {}
    if extend_names is None:
        extend_names = {}
    if extend_closures is None:
        extend_closures = {}

    # collection of {name:Symbol} mappings for `sympy.parse_expr()`
    local_dict = symbols.copy()  # dereference so extend values can be added

    # enable user-provided functions (NOTE these might be {})
    local_dict.update({s.name: s for s in extend_names.keys()})
    local_dict.update({s.name: s for s in extend_closures.keys()})

    # get indexing Symbols (IndexedBase, Idx) and the spread of any indexer(s)
    #   members of symbols will be used if they're the correct type, otherwise
    #   TypeError will be raised for names which exist, but are not valid for
    #   their respective types (see `get_or_create_symbol()`)
    # NOTE for now there can only be exactly 1 or 0 indexers (Idx) (for now?)
    offset_values, offset_ranges = string_expr_indexing_offsets(expr_string, symbols)

    # continue to build up symbols dict for `sympy.parse_expr()`
    local_dict.update(offset_values)

    # convert forms like `expr_rhs` into `Eq(result_lhs, expr_rhs)`
    verify_literal_result_symbol = False  # avoid NameError in later check
    if not expr_string.startswith("Eq("):
        if "=" in expr_string:
            raise RuntimeError(f"BUG: failed to handle equality during cleanup: {expr_string}")
        if name_result is None:
            verify_literal_result_symbol = True  # enable later warning path checks
            name_result = "result"
        # rewrite `expr_string` to `Eq()` form
        if offset_values:
            syms_result = get_or_create_symbol(symbols, name_result, sympy.IndexedBase)
            # FUTURE reconsider if supporting multiple indexers
            # unpack name (rather than smuggling it from the earlier loop..)
            indexer = next(iter(offset_ranges))
            expr_string = f"Eq({syms_result.name}[{indexer.name}], {expr_string})"
        else:
            syms_result = get_or_create_symbol(symbols, name_result, sympy.Symbol)
            expr_string = f"Eq({syms_result.name}, {expr_string})"
        # pack result into locals before parse
        local_dict.update({name_result: syms_result})

    expr_sympy = sympy.parse_expr(expr_string, local_dict=local_dict)

    if not expr_sympy.atoms(sympy.Eq):  # ensures (lhs,rhs) properties, alt: hasattr
        raise RuntimeError(f"BUG: didn't coerce into Eq(LHS, RHS) form: {expr_sympy}")

    # now (re-)extract the result Symbol from LHS
    # NOTE IndexedBase coerced to Symbol [ISSUE 9]
    atoms_lhs = expr_sympy.lhs.atoms(sympy.Symbol)
    # FUTURE opportunity to extract Number from LHS to fail or divide out
    if len(atoms_lhs) == 1:
        pass  # pop later, set of exactly 1 Symbol
    elif len(atoms_lhs) == 2:
        atoms_lhs = expr_sympy.lhs.atoms(sympy.IndexedBase)
        if len(atoms_lhs) != 1:
            raise ValueError(f"multiple possible result values: {atoms_lhs}")
    else:
        raise ValueError(f"multiple or no possible result values from LHS atoms:{atoms_lhs}")
    symbol_result = atoms_lhs.pop()  # now dissolve set: {x} -> x

    if name_result is not None and name_result != symbol_result.name:
        raise ValueError(f"mismatch between name_result ({name_result}) and parsed symbol name ({symbol_result.name})")

    # make dicts of {name:Symbol} for caller
    # NOTE `symbol_result` must be last to simplify dropping via slicing in later logic
    # NOTE `.atoms(Symbol)` picks up IndexedBase, but demotes them to new `Symbol` instances [ISSUE 9]

    # warn the user if they passed unused symbols
    symbols_unused = set(symbols.keys()) - {s.name for s in expr_sympy.atoms(sympy.Symbol)}
    if symbols_unused:  # should be emptyset
        warn(f"some symbols were not used: {tidy_list_str(symbols_unused)}")

    symbols = {s.name: s for s in expr_sympy.atoms(sympy.Symbol)}
    symbols.update({s.name: s for s in expr_sympy.atoms(sympy.IndexedBase)})
    symbols.pop(symbol_result.name)  # restored later as the last entry
    for indexer in offset_ranges.keys():  # expressly remove Idx name(s)
        del symbols[indexer.name]

    # force lexical ordering by-name for consistency (becomes args, etc.)
    symbols = {name: symbols[name] for name in sorted(symbols.keys())}

    # make a dict (len==1) of the result symbol
    syms_result = {symbol_result.name: symbol_result}
    # now append it to the symbols dict so it can be an argument
    symbols.update(syms_result)  # always the last symbol

    # hint that user may be misusing "result" name in their RHS
    if verify_literal_result_symbol and (
        name_result in {a.name for a in expr_sympy.rhs.atoms(sympy.Symbol)}) and (
        name_result not in offset_values.keys()
    ):
        warn("symbol 'result' in RHS refers to result array, but not indexed or passed as name_result")

    return expr_sympy, symbols, offset_ranges, syms_result


def parse_sympy_expr(expr, name_result, symbols):
    """ get a compatible set of objects for later use, mirroring `string_expr_to_sympy()`,
        but for a valid SymPy expr (may be an Expr or Equality)
            expr           SymPy expr
            symbols        {s:s.name}             sorted, no indexer, result last
            offset_ranges  {indexer:[min,max]}    indexer is Idx
            result         {result:resultsymbol}  always a dict of len 1
    """
    if symbols:  # NOTE rewritten later from expr
        symbols_unused = set(symbols.values()) - expr.atoms(sympy.Symbol, sympy.IndexedBase, sympy.Idx)
        if symbols_unused:
            raise ValueError(f"some symbols not present in expr: {symbols_unused}")

    offset_values, offset_ranges = indexed_offsets_from_expr(expr)

    if expr.atoms(sympy.Eq):  # form Eq(LHS,RHS) TODO consider `isinstance(e,Equality)` instead
        if len(expr.atoms(sympy.Eq)) != 1:
            raise ValueError(f"only a single equality can exist, but got {expr.atoms(sympy.Eq)}")
        result = (expr.lhs.atoms(sympy.IndexedBase) or expr.lhs.atoms(sympy.Symbol))
        if len(result) != 1:  # `indexed_offsets_from_expr()` ensures only a single value exists
            raise ValueError(f"BUG: expected a single result, but got {expr.lhs}")
        result = result.pop()
        if name_result is not None and result.name != name_result:
            raise ValueError(f"mismatched name between name_result({name_result}) and LHS({result})")
    else:  # form RHS -> Eq(result,RHS)
        # NOTE because all symbols exist, user can't pass "result" naively
        if name_result is None:
            name_result = "result"
            if "result" in (a.name for a in expr.atoms(sympy.Symbol, sympy.IndexedBase, sympy.Idx)):
                warn("symbol 'result' in RHS refers to result array, but not indexed or passed as name_result")
        if expr.atoms(sympy.IndexedBase):  # RHS reveals indexing
            indexer = next(iter(offset_ranges))
            result  = get_or_create_symbol(symbols, name_result, sympy.IndexedBase)
            expr    = sympy.Eq(result[indexer], expr)
        else:
            result  = get_or_create_symbol(symbols, name_result, sympy.Symbol)
            expr    = sympy.Eq(result, expr)

    symbols = {s.name: s for s in expr.atoms(sympy.Symbol)}
    symbols.update({s.name: s for s in expr.atoms(sympy.IndexedBase)})
    symbols.pop(result.name)  # restored later as the last entry
    for indexer in offset_ranges.keys():  # expressly remove Idx name(s)
        del symbols[indexer.name]

    # force lexical ordering by-name for consistency (becomes args, etc.)
    symbols = {name: symbols[name] for name in sorted(symbols.keys())}

    # make a dict (len==1) of the result symbol
    result_dict = {result.name: result}
    # now append it to the symbols dict so it can be an argument
    symbols.update(result_dict)  # always the last symbol

    return expr, symbols, offset_ranges, result_dict


def parse_inputs(expr, symbols, name_result, extend):
    # clean up initial symbols
    symbols = symbols_given_cleanup(expr, symbols)

    # TODO more aggressively insist it's a valid Python name
    if name_result is not None and not isinstance(name_result, str):
        raise ValueError(f"name_result must be None or a str, but got {type(name_result)}")

    # split extend into {SymbolOptional: number} and {sympy.Function: string function}
    extend_names, extend_closures = extra_functionality_parse(extend)

    if isinstance(expr, str):
        expr = string_expr_cleanup(expr)
        expr, symbols, indexers, results = string_expr_to_sympy(expr, name_result, symbols, extend_names, extend_closures)
    elif isinstance(expr, (sympy.core.expr.Expr, sympy.core.relational.Equality)):
        expr, symbols, indexers, results = parse_sympy_expr(expr, name_result, symbols)
        # TODO detect non-standard Functions in expr which are not in the extend collections?
        #   maybe `.free_symbols` can help?
    else:
        raise ValueError(f"unexpected expr type({type(expr)}), must be str or SymPy Expr")

    # take any dummy symbols out of collection as they're not required in data
    symbols, symbols_dummy = dummy_symbols_split(expr, symbols)

    return expr, symbols, indexers, results, symbols_dummy, extend_names, extend_closures


def parse_parallel_model(parallel_model=None):
    # TODO should this be merged with instance config? probably yes

    # keep defaults (subject to change, likely warning for non-None)
    if parallel_model is None:
        return {}

    # expressly disable all parallization
    # won't lose errors in Numba threads and slightly-shorter build times
    if parallel_model == "disabled":
        return {
            "builder.parallel_model.numba_parallel":    False,
            "builder.parallel_model.allow_prange":      False,
            "builder.parallel_model.native_threadpool": False,
        }

    if parallel_model == "prange":
        return {
            "builder.parallel_model.numba_parallel": True,
            "builder.parallel_model.allow_prange":   True,
        }

    if parallel_model == "native_threadpool":
        return {
            "builder.parallel_model.numba_parallel":    False,
            "builder.parallel_model.allow_prange":      False,
            "builder.parallel_model.native_threadpool": True,
        }

    raise ValueError(f"parallel_model must be one of (None,'disabled','prange'), but got '{parallel_model}'")
